"""
mathematical functions for complex numbers.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/cmath.rst.
==========================================================

.. module:: cmath
   :synopsis: mathematical functions for complex numbers

|see_cpython_module| :mod:`python:cmath`.

The ``cmath`` module provides some basic mathematical functions for
working with complex numbers.

Availability: not available on WiPy and ESP8266. Floating point support
required for this module.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import SupportsComplex, SupportsFloat, Final

_C: Final = SupportsFloat | SupportsComplex

def cos(z: _C, /) -> complex:
    """
   Return the cosine of ``z``.
   """

def exp(z: _C, /) -> complex:
    """
   Return the exponential of ``z``.
   """

def log(z: _C, /) -> complex:
    """
   Return the natural logarithm of ``z``.  The branch cut is along the negative real axis.
   """

def log10(z: _C, /) -> complex:
    """
   Return the base-10 logarithm of ``z``.  The branch cut is along the negative real axis.
   """

def phase(z: _C, /) -> float:
    """
   Returns the phase of the number ``z``, in the range (-pi, +pi].
   """

def polar(z: _C, /) -> tuple[float, float]:
    """
   Returns, as a tuple, the polar form of ``z``.
   """

def rect(r: float, phi: float, /) -> complex:
    """
   Returns the complex number with modulus ``r`` and phase ``phi``.
   """

def sin(z: _C, /) -> complex:
    """
   Return the sine of ``z``.
   """

def sqrt(z: _C, /) -> complex:
    """
   Return the square-root of ``z``.
   """

e: Final[float] = ...
"""
base of the natural logarithm
"""

pi: Final[float] = ...
"""
the ratio of a circle's circumference to its diameter
"""



# -*- coding: utf-8 -*-
import base64
import datetime
import json
import os.path
import re
import sys
import tkinter as tk
import platform
import traceback
import urllib
from logging import exception, getLogger
from tkinter import messagebox, simpledialog, ttk
from typing import Optional, Union
import cv2
import requests
from _tkinter import TclError

from turcar import get_runner, get_workbench, ui_utils
from turcar.base_file_browser import ask_backend_path, choose_node_for_file_operations
from turcar.codeview import BinaryFileException, CodeView, CodeViewText
from turcar.common import (
    REMOTE_PATH_MARKER,
    InlineCommand,
    TextRange,
    ToplevelResponse,
    is_local_path,
    is_remote_path,
    is_same_path,
    normpath_with_actual_case,
    universal_dirname,
)
from turcar.custom_notebook import CustomNotebook, CustomNotebookTab
from turcar.languages import tr
from turcar.misc_utils import running_on_mac_os, running_on_windows
from turcar.tktextext import rebind_control_a
from turcar.ui_utils import (
    askopenfilename,
    asksaveasfilename,
    get_beam_cursor,
    select_sequence,
    windows_known_extensions_are_hidden,
)
from turcar.video import VideoDialog

PYTHON_FILES_STR = tr("Python files")
_dialog_filetypes = [(PYTHON_FILES_STR, ".py .pyw .pyi"), (tr("all files"), ".*")]

PYTHON_EXTENSIONS = {"py", "pyw", "pyi"}
PYTHONLIKE_EXTENSIONS = set()

logger = getLogger(__name__)


class EditorCodeViewText(CodeViewText):
    """Allows separate class binding for CodeViewTexts which are inside editors"""

    def __init__(self, master=None, cnf={}, **kw):
        super().__init__(
            master=master,
            cnf=cnf,
            **kw,
        )
        self.bindtags(self.bindtags() + ("EditorCodeViewText",))


class Editor(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master)
        assert isinstance(master, EditorNotebook)
        self.containing_notebook = master  # type: EditorNotebook

        self._code_view = CodeView(
            self,
            propose_remove_line_numbers=True,
            font="EditorFont",
            text_class=EditorCodeViewText,
            cursor=get_beam_cursor(),
        )
        get_workbench().event_generate(
            "EditorTextCreated", editor=self, text_widget=self.get_text_widget()
        )

        self._code_view.grid(row=0, column=0, sticky=tk.NSEW, in_=self)

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self._filename = None
        self._last_known_mtime = None

        self._code_view.text.bind("<<Modified>>", self._on_text_modified, True)
        self._code_view.text.bind("<<TextChange>>", self._on_text_change, True)
        self._code_view.text.bind("<Control-Tab>", self._control_tab, True)

        get_workbench().bind("DebuggerResponse", self._listen_debugger_progress, True)
        get_workbench().bind("ToplevelResponse", self._listen_for_toplevel_response, True)

        self.update_appearance()

    def get_text_widget(self):
        return self._code_view.text

    def get_code_view(self):
        # TODO: try to get rid of this
        return self._code_view

    def get_content(self) -> str:
        return self._code_view.get_content()

    def set_filename(self, path):
        self._filename = path

    def get_filename(self, try_hard=False):
        if self._filename is None and try_hard:
            self.save_file()

        return self._filename

    def get_identifier(self):
        if self._filename:
            return self._filename
        else:
            return str(self.winfo_id())

    def get_title(self):
        if self.get_filename() is None:
            result = tr("<untitled>")
        elif is_remote_path(self.get_filename()):
            path = extract_target_path(self.get_filename())
            name = path.split("/")[-1]
            result = "[ " + name + " ]"
        else:
            result = os.path.basename(self.get_filename())

        if self.is_modified():
            result += " *"

        return result

    def check_for_external_changes(self):
        if self._filename is None:
            return

        if is_remote_path(self._filename):
            return

        if self._last_known_mtime is None:
            return

        elif not os.path.exists(self._filename):
            self.master.select(self)

            if messagebox.askyesno(
                    tr("File is gone"),
                    tr("Looks like '%s' was deleted or moved.") % self._filename
                    + "\n\n"
                    + tr("Do you want to also close the editor?"),
                    master=self,
            ):
                self.master.close_editor(self)
            else:
                self.get_text_widget().edit_modified(True)
                self._last_known_mtime = None

        elif os.path.getmtime(self._filename) != self._last_known_mtime:
            skip_confirmation = not self.is_modified() and get_workbench().get_option(
                "edit.auto_refresh_saved_files"
            )
            if not skip_confirmation:
                self.master.select(self)

            if skip_confirmation or messagebox.askyesno(
                    tr("External modification"),
                    tr("Looks like '%s' was modified outside of the editor.") % self._filename
                    + "\n\n"
                    + tr(
                        "Do you want to discard current editor content and reload the file from disk?"
                    ),
                    master=self,
            ):
                prev_location = self.get_text_widget().index("insert")
                self._load_file(self._filename, keep_undo=True)
                try:
                    self.get_text_widget().mark_set("insert", prev_location)
                    self.see_line(int(prev_location.split(".")[0]))
                except Exception:
                    logger.exception("Could not restore previous location")

            self._last_known_mtime = os.path.getmtime(self._filename)

    def get_long_description(self):
        if self._filename is None:
            result = tr("<untitled>")
        else:
            result = self._filename

        try:
            index = self._code_view.text.index("insert")
            if index and "." in index:
                line, col = index.split(".")
                result += "  @  {} : {}".format(line, int(col) + 1)
        except Exception:
            exception("Finding cursor location")

        return result

    def _load_file(self, filename, keep_undo=False):
        try:
            if is_remote_path(filename):
                result = self._load_remote_file(filename)
            else:
                result = self._load_local_file(filename, keep_undo)
                if not result:
                    return False
        except BinaryFileException:
            messagebox.showerror(
                tr("Problem"), tr("%s doesn't look like a text file") % (filename,), master=self
            )
            return False
        except SyntaxError as e:
            assert "encoding" in str(e).lower()
            messagebox.showerror(
                tr("Problem loading file"),
                tr(
                    "This file seems to have problems with encoding.\n\n"
                    "Make sure it is in UTF-8 or contains proper encoding hint."
                ),
                master=self,
            )
            return False

        self.update_appearance()
        return True

    def _load_local_file(self, filename, keep_undo=False):
        with open(filename, "rb") as fp:
            source = fp.read()

        # Make sure Windows filenames have proper format
        filename = normpath_with_actual_case(filename)
        self._filename = filename
        self.update_file_type()
        self._last_known_mtime = os.path.getmtime(self._filename)

        get_workbench().event_generate("Open", editor=self, filename=filename)
        if not self._code_view.set_content_as_bytes(source, keep_undo):
            return False
        self.get_text_widget().edit_modified(False)
        self._code_view.focus_set()
        self.master.remember_recent_file(filename)
        return True

    def _load_remote_file(self, filename):
        self._filename = filename
        self.update_file_type()
        self._code_view.set_content("")
        self._code_view.text.set_read_only(True)

        target_filename = extract_target_path(self._filename)

        self.update_title()
        response = get_runner().send_command_and_wait(
            InlineCommand(
                "read_file", path=target_filename, description=tr("Loading %s") % target_filename
            ),
            dialog_title=tr("Loading"),
        )

        if response.get("error"):
            # TODO: make it softer
            raise RuntimeError(response["error"])

        content = response["content_bytes"]
        self._code_view.text.set_read_only(False)
        if not self._code_view.set_content_as_bytes(content):
            return False
        self.get_text_widget().edit_modified(False)
        self.update_title()
        return True

    def is_modified(self):
        return bool(self._code_view.text.edit_modified())

    def save_file_enabled(self):
        return self.is_modified() or not self.get_filename()

    def save_file(self, ask_filename=False, save_copy=False, node=None) -> Optional[str]:
        if self._filename is not None and not ask_filename:
            save_filename = self._filename
            get_workbench().event_generate("Save", editor=self, filename=save_filename)
        else:
            save_filename = self.ask_new_path(node)

            if not save_filename:
                return None

            if self.containing_notebook.get_editor(save_filename) is not None:
                messagebox.showerror(
                    tr("File is open"),
                    tr(
                        "This file is already open in turcar.\n\n"
                        "If you want to save with this name,\n"
                        "close the existing editor first!"
                    ),
                    master=get_workbench(),
                )
                return None

            get_workbench().event_generate(
                "SaveAs", editor=self, filename=save_filename, save_copy=save_copy
            )

        content_bytes = self._code_view.get_content_as_bytes()

        if is_remote_path(save_filename):
            result = self.write_remote_file(save_filename, content_bytes, save_copy)
        else:
            result = self.write_local_file(save_filename, content_bytes, save_copy)

        if not result:
            return None

        if not save_copy:
            self._filename = save_filename
            self.update_file_type()

        self.update_title()
        return save_filename

    def write_local_file(self, save_filename, content_bytes, save_copy):
        process_shebang = content_bytes.startswith(b"#!/") and get_workbench().get_option(
            "file.make_saved_shebang_scripts_executable"
        )
        if process_shebang:
            content_bytes = content_bytes.replace(b"\r\n", b"\n")

        try:
            f = open(save_filename, mode="wb")
            f.write(content_bytes)
            f.flush()
            # Force writes on disk, see https://learn.adafruit.com/adafruit-circuit-playground-express/creating-and-editing-code#1-use-an-editor-that-writes-out-the-file-completely-when-you-save-it
            os.fsync(f)
            f.close()
            if process_shebang:
                os.chmod(save_filename, 0o755)
            if not save_copy or save_filename == self._filename:
                self._last_known_mtime = os.path.getmtime(save_filename)
            get_workbench().event_generate(
                "LocalFileOperation", path=save_filename, operation="save"
            )
        except PermissionError:
            messagebox.showerror(
                tr("Permission Error"),
                tr("Looks like this file or folder is not writable."),
                master=self,
            )
            return False

        if not save_copy or save_filename == self._filename:
            self.master.remember_recent_file(save_filename)

        if not save_copy or save_filename == self._filename:
            self._code_view.text.edit_modified(False)

        return True

    def update_file_type(self):
        if self._filename is None:
            self._code_view.set_file_type(None)
        else:
            ext = self._filename.split(".")[-1].lower()
            if ext in PYTHON_EXTENSIONS:
                file_type = "python"
            elif ext in PYTHONLIKE_EXTENSIONS:
                file_type = "pythonlike"
            else:
                file_type = None

            self._code_view.set_file_type(file_type)

        self.update_appearance()

    def write_remote_file(self, save_filename, content_bytes, save_copy):
        if get_runner().ready_for_remote_file_operations(show_message=True):
            target_filename = extract_target_path(save_filename)

            result = get_runner().send_command_and_wait(
                InlineCommand(
                    "write_file",
                    path=target_filename,
                    content_bytes=content_bytes,
                    editor_id=id(self),
                    blocking=True,
                    description=tr("Saving to %s") % target_filename,
                    make_shebang_scripts_executable=get_workbench().get_option(
                        "file.make_saved_shebang_scripts_executable"
                    ),
                ),
                dialog_title=tr("Saving"),
            )

            if "error" in result:
                messagebox.showerror(tr("Could not save"), str(result["error"]))
                return False

            if not save_copy:
                self._code_view.text.edit_modified(False)

            self.update_title()

            # NB! edit_modified is not falsed yet!
            get_workbench().event_generate(
                "RemoteFileOperation", path=target_filename, operation="save"
            )
            get_workbench().event_generate("RemoteFilesChanged")
            return True
        else:
            messagebox.showerror(tr("Could not save"), tr("Back-end is not ready"))
            return False

    def ask_new_path(self, node=None):
        if node is None:
            node = choose_node_for_file_operations(self.winfo_toplevel(), tr("Where to save to?"))
        if not node:
            return None

        if node == "local":
            return self.ask_new_local_path()
        else:
            assert node == "remote"
            return self.ask_new_remote_path()

    def ask_new_remote_path(self):
        target_path = ask_backend_path(self.winfo_toplevel(), "save")
        if target_path:
            target_path = self._check_add_py_extension(target_path)
            return make_remote_path(target_path)
        else:
            return None

    def ask_new_local_path(self):
        if self._filename is None:
            initialdir = get_workbench().get_local_cwd()
            initialfile = None
        else:
            initialdir = os.path.dirname(self._filename)
            initialfile = os.path.basename(self._filename)

        # https://tcl.tk/man/tcl8.6/TkCmd/getOpenFile.htm
        type_var = tk.StringVar(value="")
        new_filename = asksaveasfilename(
            filetypes=_dialog_filetypes,
            defaultextension=None,
            initialdir=initialdir,
            initialfile=initialfile,
            parent=get_workbench(),
            typevariable=type_var,
        )
        logger.info("Save dialog returned %r with typevariable %r", new_filename, type_var.get())

        # Different tkinter versions may return different values
        if new_filename in ["", (), None]:
            return None

        if running_on_windows():
            # may have /-s instead of \-s and wrong case
            new_filename = os.path.join(
                normpath_with_actual_case(os.path.dirname(new_filename)),
                os.path.basename(new_filename),
            )

        if type_var.get() == PYTHON_FILES_STR or type_var.get() == "":
            new_filename = self._check_add_py_extension(
                new_filename, without_asking=type_var.get() == PYTHON_FILES_STR
            )

        if new_filename.endswith(".py"):
            base = os.path.basename(new_filename)
            mod_name = base[:-3].lower()
            if running_on_windows():
                mod_name = mod_name.lower()

            if mod_name in [
                "math",
                "turtle",
                "random",
                "statistics",
                "pygame",
                "matplotlib",
                "numpy",
            ]:
                # More proper name analysis will be performed by ProgramNamingAnalyzer
                if not tk.messagebox.askyesno(
                        tr("Potential problem"),
                        tr(
                            "If you name your script '%s', "
                            "you won't be able to import the library module named '%s'"
                        )
                        % (base, mod_name)
                        + ".\n\n"
                        + tr("Do you still want to use this name for your script?"),
                        master=self,
                ):
                    return self.ask_new_local_path()

        return new_filename

    def show(self):
        self.master.select(self)

    def update_appearance(self):
        self._code_view.set_gutter_visibility(
            get_workbench().get_option("view.show_line_numbers") or get_workbench().in_simple_mode()
        )
        self._code_view.set_line_length_margin(
            get_workbench().get_option("view.recommended_line_length")
        )
        self._code_view.text.update_tabs()
        self._code_view.text.event_generate("<<UpdateAppearance>>")
        self._code_view.grid_main_widgets()

    def _listen_debugger_progress(self, event):
        # Go read-only
        # TODO: check whether this module is active?
        self._code_view.text.set_read_only(True)

    def _listen_for_toplevel_response(self, event: ToplevelResponse) -> None:
        self._code_view.text.set_read_only(False)

    def _control_tab(self, event):
        if event.state & 1:  # shift was pressed
            direction = -1
        else:
            direction = 1
        self.master.select_next_prev_editor(direction)
        return "break"

    def _shift_control_tab(self, event):
        self.master.select_next_prev_editor(-1)
        return "break"

    def select_range(self, text_range):
        self._code_view.select_range(text_range)

    def select_line(self, lineno, col_offset=None):
        self._code_view.select_range(TextRange(lineno, 0, lineno + 1, 0))
        self.see_line(lineno)

        if col_offset is None:
            col_offset = 0

        self.get_text_widget().mark_set("insert", "%d.%d" % (lineno, col_offset))

    def see_line(self, lineno):
        # first see an earlier line in order to push target line downwards
        self._code_view.text.see("%d.0" % max(lineno - 4, 1))
        self._code_view.text.see("%d.0" % lineno)

    def focus_set(self):
        self._code_view.focus_set()

    def is_focused(self):
        return self.focus_displayof() == self._code_view.text

    def _on_text_modified(self, event):
        self.update_title()

    def update_title(self):
        try:
            self.master.update_editor_title(self)
        except Exception:
            logger.exception("Could not update editor title")

    def _on_text_change(self, event):
        # may not be added to the Notebook yet
        if self.containing_notebook.has_content(self):
            self.update_title()

    def destroy(self):
        get_workbench().unbind("DebuggerResponse", self._listen_debugger_progress)
        get_workbench().unbind("ToplevelResponse", self._listen_for_toplevel_response)
        ttk.Frame.destroy(self)
        get_workbench().event_generate(
            "EditorTextDestroyed", editor=self, text_widget=self.get_text_widget()
        )

    def _check_add_py_extension(self, path: str, without_asking: bool = False) -> str:
        assert path
        parts = re.split(r"[/\\]", path)
        name = parts[-1]
        if "." not in name:
            if without_asking or messagebox.askyesno(
                    title=tr("Confirmation"),
                    message=tr("Python files usually have .py extension.")
                            + "\n\n"
                            + tr("Did you mean '%s'?" % (name + ".py")),
                    parent=self.winfo_toplevel(),
            ):
                return path + ".py"
            else:
                return path

        return path


class EditorNotebook(CustomNotebook):
    """
    Manages opened files / modules
    """

    def __init__(self, master):
        super().__init__(master)

        get_workbench().set_default("file.reopen_all_files", False)
        get_workbench().set_default("file.open_files", [])
        get_workbench().set_default("file.current_file", None)
        get_workbench().set_default("file.recent_files", [])
        get_workbench().set_default("view.highlight_current_line", False)
        get_workbench().set_default("view.show_line_numbers", True)
        get_workbench().set_default("view.recommended_line_length", 0)
        get_workbench().set_default("edit.indent_with_tabs", False)
        get_workbench().set_default("edit.auto_refresh_saved_files", True)
        get_workbench().set_default("file.make_saved_shebang_scripts_executable", True)

        self._recent_menu = tk.Menu(
            get_workbench().get_menu("file"), postcommand=self._update_recent_menu
        )
        self._init_commands()
        self.enable_traversal()

        # open files from last session
        """ TODO: they should go only to recent files
        for filename in prefs["open_files"].split(";"):
            if os.path.exists(filename):
                self._open_file(filename)
        """

        # should be in the end, so that it can be detected when
        # constructor hasn't completed yet
        self.preferred_size_in_pw = None
        self._checking_external_changes = False

        get_workbench().bind("WindowFocusIn", self.check_for_external_changes, True)
        get_workbench().bind("ToplevelResponse", self.check_for_external_changes, True)
        self.bind("<<NotebookTabChanged>>", self.on_tab_changed, True)

    def on_tab_changed(self, *args):
        # Required to avoid incorrect sizing of parent panes
        self.update_idletasks()

    def _init_commands(self):
        # TODO: do these commands have to be in EditorNotebook ??
        # Create a module level function install_editor_notebook ??
        # Maybe add them separately, when notebook has been installed ??

        get_workbench().add_command(
            "new_file",
            "file",
            tr("New"),
            self._cmd_new_file,
            caption=tr("New"),
            default_sequence=select_sequence("<Control-n>", "<Command-n>"),
            extra_sequences=["<Control-Greek_nu>"],
            group=10,
            image="new-file",
            include_in_toolbar=True,
        )

        get_workbench().add_command(
            "open_file",
            "file",
            tr("Open..."),
            self._cmd_open_file,
            caption=tr("Load"),
            default_sequence=select_sequence("<Control-o>", "<Command-o>"),
            extra_sequences=["<Control-Greek_omicron>"],
            group=10,
            image="open-file",
            include_in_toolbar=True,
        )

        get_workbench().add_command(
            "recents", "file", tr("Recent files"), group=10, submenu=self._recent_menu
        )

        # http://stackoverflow.com/questions/22907200/remap-default-keybinding-in-tkinter
        get_workbench().bind_class("Text", "<Control-o>", self._control_o)
        get_workbench().bind_class("Text", "<Control-Greek_omicron>", self._control_o)
        rebind_control_a(get_workbench())

        get_workbench().add_command(
            "close_file",
            "file",
            tr("Close"),
            self._cmd_close_file,
            default_sequence=select_sequence("<Control-w>", "<Command-w>"),
            extra_sequences=["<Control-Greek_finalsmallsigma>"],
            tester=lambda: self.get_current_editor() is not None,
            group=10,
        )

        get_workbench().add_command(
            "close_files",
            "file",
            tr("Close all"),
            self.close_tabs,
            tester=lambda: self.get_current_editor() is not None,
            default_sequence=select_sequence("<Control-W>", "<Command-Alt-w>"),
            group=10,
        )

        get_workbench().add_command(
            "save_file",
            "file",
            tr("Save"),
            self._cmd_save_file,
            caption=tr("Save"),
            default_sequence=select_sequence("<Control-s>", "<Command-s>"),
            extra_sequences=["<Control-Greek_sigma>"],
            tester=self._cmd_save_file_enabled,
            group=10,
            image="save-file",
            include_in_toolbar=True,
        )

        get_workbench().add_command(
            "save_all_files",
            "file",
            tr("Save All files"),
            self._cmd_save_all_files,
            caption=tr("Save All files"),
            default_sequence=select_sequence("<Control-Alt-s>", "<Command-Alt-s>"),
            tester=self._cmd_save_all_files_enabled,
            group=10,
        )

        get_workbench().add_command(
            "save_file_as",
            "file",
            tr("Save as..."),
            self._cmd_save_file_as,
            default_sequence=select_sequence("<Control-Shift-S>", "<Command-Shift-S>"),
            extra_sequences=["<Control-Greek_SIGMA>"],
            tester=lambda: self.get_current_editor() is not None,
            group=10,
        )

        get_workbench().add_command(
            "save_copy",
            "file",
            tr("Save copy..."),
            self._cmd_save_copy,
            tester=lambda: self.get_current_editor() is not None,
            group=10,
        )

        get_workbench().add_command(
            "move_rename_file",
            "file",
            tr("Move / rename..."),
            self._cmd_move_rename_file,
            tester=self._cmd_move_rename_file_enabled,
            group=10,
        )

        get_workbench().add_command(
            "goto_source_line",
            "edit",
            tr("Go to line..."),
            self._cmd_goto_source_line,
            default_sequence=select_sequence("<Control-g>", "<Command-g>"),
            # tester=,
            # no global switch, or cross plugin switch?
            # todo use same as find and replace -> plugins/find_replace.py
            group=60,
        )

        get_workbench().add_command(
            "watch_video",
            "course",
            tr("课程视频"),
            self._cmd_watch_video,
            caption=tr("Video"),
            group=10,
            image="video",
            include_in_toolbar=True,
        )

        get_workbench().add_command(
            "face_recognition",
            "recognition",
            tr("百度人脸识别"),
            self._cmd_face_recognition,
            caption=tr("Recognition"),
            group=10,
            image="face_recog",
            include_in_toolbar=True,
        )

        get_workbench().createcommand("::tk::mac::OpenDocument", self._mac_open_document)

    def load_startup_files(self):
        """If no filename was sent from command line
        then load previous files (if setting allows)"""

        cmd_line_filenames = [
            os.path.abspath(name) for name in sys.argv[1:] if os.path.exists(name)
        ]

        if len(cmd_line_filenames) > 0:
            filenames = cmd_line_filenames
        elif get_workbench().get_option("file.reopen_all_files"):
            filenames = get_workbench().get_option("file.open_files")
        elif get_workbench().get_option("file.current_file"):
            filenames = [get_workbench().get_option("file.current_file")]
        else:
            filenames = []

        if len(filenames) > 0:
            for filename in filenames:
                if os.path.exists(filename):
                    self.show_file(filename)

            cur_file = get_workbench().get_option("file.current_file")
            # choose correct active file
            if len(cmd_line_filenames) > 0:
                self.show_file(cmd_line_filenames[0])
            elif cur_file and os.path.exists(cur_file):
                self.show_file(cur_file)
            else:
                self._cmd_new_file()
        else:
            self._cmd_new_file()

    def save_all_named_editors(self):
        all_saved = True
        for editor in self.winfo_children():
            if editor.get_filename() and editor.is_modified():
                success = editor.save_file()
                all_saved = all_saved and success

        return all_saved

    def remember_recent_file(self, filename):
        recents = get_workbench().get_option("file.recent_files")
        if filename in recents:
            recents.remove(filename)
        recents.insert(0, filename)
        relevant_recents = [name for name in recents if os.path.exists(name)][:15]
        get_workbench().set_option("file.recent_files", relevant_recents)
        self._update_recent_menu()

    def _update_recent_menu(self):
        recents = get_workbench().get_option("file.recent_files")
        relevant_recents = [
            path for path in recents if os.path.exists(path) and not self.file_is_opened(path)
        ]
        self._recent_menu.delete(0, "end")
        for path in relevant_recents:
            def load(path=path):
                self.show_file(path)

            self._recent_menu.insert_command("end", label=path, command=load)

    def remember_open_files(self):
        if (
                self.get_current_editor() is not None
                and self.get_current_editor().get_filename() is not None
        ):
            current_file = self.get_current_editor().get_filename()
        else:
            current_file = None

        get_workbench().set_option("file.current_file", current_file)

        open_files = [
            editor.get_filename() for editor in self.winfo_children() if editor.get_filename()
        ]
        get_workbench().set_option("file.open_files", open_files)

    def _cmd_new_file(self):
        self.open_new_file()

    def open_new_file(self, path=None, remote=False):
        new_editor = Editor(self)
        get_workbench().event_generate("NewFile", editor=new_editor)
        if path:
            if remote:
                new_editor.set_filename(make_remote_path(path))
            else:
                new_editor.set_filename(path)
        self.add(new_editor, text=new_editor.get_title())
        self.select(new_editor)
        new_editor.focus_set()

    def _cmd_open_file(self):
        node = choose_node_for_file_operations(self.winfo_toplevel(), "Where to open from?")
        if not node:
            return

        if node == "local":
            initialdir = get_workbench().get_local_cwd()
            if (
                    self.get_current_editor() is not None
                    and self.get_current_editor().get_filename() is not None
            ):
                initialdir = os.path.dirname(self.get_current_editor().get_filename())
            path = askopenfilename(
                filetypes=_dialog_filetypes, initialdir=initialdir, parent=get_workbench()
            )
        else:
            assert node == "remote"
            target_path = ask_backend_path(self.winfo_toplevel(), "open")
            if not target_path:
                return

            path = make_remote_path(target_path)

        if path:
            # self.close_single_untitled_unmodified_editor()
            self.show_file(path, propose_dialog=False)

    def _control_o(self, event):
        # http://stackoverflow.com/questions/22907200/remap-default-keybinding-in-tkinter
        self._cmd_open_file()
        return "break"

    def _close_files(self, except_index=None):
        for tab_index in reversed(range(len(self.winfo_children()))):
            if except_index is not None and tab_index == except_index:
                continue
            else:
                editor = self.get_child_by_index(tab_index)
                self.close_editor(editor, force=False)

    def _cmd_close_file(self):
        self.close_tab(self.index(self.select()))

    def close_tab(self, index_or_tab: Union[int, CustomNotebookTab]):
        if isinstance(index_or_tab, int):
            page = self.pages[index_or_tab]
        else:
            page = self.get_page_by_tab(index_or_tab)

        assert isinstance(page.content, Editor)
        self.close_editor(page.content)

    def close_editor(self, editor, force=False):
        if not force and not self.check_allow_closing(editor):
            return
        self._plain_forget(editor)
        editor.destroy()

    def _cmd_save_file(self):
        if self.get_current_editor():
            self.get_current_editor().save_file()
            self.update_editor_title(self.get_current_editor())

    def _cmd_save_file_enabled(self):
        return self.get_current_editor() and self.get_current_editor().save_file_enabled()

    def _cmd_save_all_files(self):
        for editor in self.get_all_editors():
            if editor.save_file_enabled() == True:
                editor.save_file()
                self.update_editor_title(editor)

    def _cmd_save_all_files_enabled(self):
        for editor in self.get_all_editors():
            if editor.save_file_enabled() == True:
                return True
        return False

    def _cmd_save_file_as(self, node=None):
        if not self.get_current_editor():
            return

        self.get_current_editor().save_file(ask_filename=True, node=node)
        self.update_editor_title(self.get_current_editor())
        get_workbench().update_title()

    def _cmd_save_copy(self):
        if not self.get_current_editor():
            return

        self.get_current_editor().save_file(ask_filename=True, save_copy=True)
        self.update_editor_title(self.get_current_editor())

    def _cmd_save_file_as_enabled(self):
        return self.get_current_editor() is not None

    def _cmd_move_rename_file(self):
        editor = self.get_current_editor()
        old_filename = editor.get_filename()
        assert old_filename is not None

        if is_remote_path(old_filename):
            node = "remote"
        else:
            node = "local"

        self._cmd_save_file_as(node=node)

        if editor.get_filename() != old_filename:
            if is_remote_path(old_filename):
                remote_path = extract_target_path(old_filename)
                get_runner().send_command_and_wait(
                    InlineCommand(
                        "delete", paths=[remote_path], description=tr("Deleting" + remote_path)
                    ),
                    dialog_title=tr("Deleting"),
                )
                get_workbench().event_generate(
                    "RemoteFileOperation", path=remote_path, operation="delete"
                )
            else:
                os.remove(old_filename)
                get_workbench().event_generate(
                    "LocalFileOperation", path=old_filename, operation="delete"
                )

    def _cmd_move_rename_file_enabled(self):
        return self.get_current_editor() and self.get_current_editor().get_filename() is not None

    def close_single_untitled_unmodified_editor(self):
        editors = self.winfo_children()
        if len(editors) == 1 and not editors[0].is_modified() and not editors[0].get_filename():
            self._cmd_close_file()

    def _cmd_goto_source_line(self):
        editor = self.get_current_editor()
        if editor:
            line_no = simpledialog.askinteger(tr("Go to line"), tr("Line number"))
            if line_no:
                editor.select_line(line_no)

    def _mac_open_document(self, *args):
        for arg in args:
            if isinstance(arg, str) and os.path.exists(arg):
                self.show_file(arg)
        get_workbench().become_active_window()

    def get_current_editor(self) -> Optional[Editor]:
        return self.get_current_child()

    def get_current_editor_content(self):
        editor = self.get_current_editor()
        if editor is None:
            return None
        else:
            return editor.get_content()

    def get_all_editors(self):
        # When workspace is closing, self.winfo_children()
        # may return an unexplainable tkinter.Frame
        return [child for child in self.winfo_children() if isinstance(child, Editor)]

    def select_next_prev_editor(self, direction):
        cur_index = self.index(self.select())
        next_index = (cur_index + direction) % len(self.tabs())
        self.select(self.get_child_by_index(next_index))

    def file_is_opened(self, path):
        for editor in self.get_all_editors():
            if editor.get_filename() and is_same_path(path, editor.get_filename()):
                return True

        return False

    def show_file(self, filename, text_range=None, set_focus=True, propose_dialog=True):
        # self.close_single_untitled_unmodified_editor()
        try:
            editor = self.get_editor(filename, True)
        except PermissionError:
            logger.exception("Loading " + filename)
            msg = tr("Got permission error when trying to load\n%s") % (filename,)
            if running_on_mac_os() and propose_dialog:
                msg += "\n\n" + tr("Try opening it with File => Open.")

            messagebox.showerror(tr("Permission error"), msg, master=self)
            return None

        if editor is None:
            return

        self.select(editor)
        if set_focus:
            editor.focus_set()

        if text_range is not None:
            editor.select_range(text_range)

        return editor

    def show_remote_file(self, target_filename):
        if not get_runner().ready_for_remote_file_operations(show_message=True):
            return None
        else:
            return self.show_file(make_remote_path(target_filename))

    # 获取视频
    def _cmd_watch_video(self):
        # print("sss")
        ui_utils.show_dialog(VideoDialog(get_workbench()))

    # 人脸识别
    def _cmd_face_recognition(self):
        # 获取当前操作系统
        current_os = platform.system()

        if current_os == "Windows":
            save_path = "D:/copy.png"
            cap = cv2.VideoCapture(0, cv2.CAP_DSHOW)  # 打开笔记本内置摄像头
        if current_os == "Darwin":  # 默认为Ubuntu或其他Linux发行版
            save_path = "turcar/copy.png"
            cap = cv2.VideoCapture(0)
        else:
            save_path = "turcar/copy.png"
            cap = cv2.VideoCapture(0)
            # todo Ubuntu摄像头获取失败
            # pass

        while cap.isOpened():  # 笔记本内置摄像头被打开后
            ret, frame = cap.read()  # 从摄像头中实时读取视频
            cv2.imshow("Recognition", frame)  # 在窗口中显示视频
            k = cv2.waitKey(1)  # 图像的刷新时间为1毫秒
            if k == ord(' '):  # 按下空格键
                cv2.imshow('img', frame)  # 显示按下空格键时摄像头视频中的图像
                cv2.imwrite(save_path, frame)  # 保存按下空格键时摄像头视频中的图像
                cap.release()  # 关闭笔记本内置摄像头
                cv2.destroyWindow("Recognition")  # 销毁名为Video的窗口
                break
        cv2.destroyAllWindows()  # 销毁显示图像的窗口
        url = "https://aip.baidubce.com/rest/2.0/face/v3/detect?access_token=" + get_access_token()

        # image 可以通过 get_file_content_as_base64("C:\fakepath\多人二维码.jpg",False) 方法获取
        # path = r"D:/copy.png"
        image = get_file_content_as_base64(save_path, False)
        payload = json.dumps({
            "image": image,
            "image_type": "BASE64",
            "face_field": "age,expression,face_shape,gender,glasses,emotion,mask",
            "max_face_num": 100
        })
        headers = {
            'Content-Type': 'application/json'
        }
        response = requests.request("POST", url, headers=headers, data=payload)
        # 解析response为JSON格式
        response_data = json.loads(response.text)
        error_code = response_data["error_code"]
        timestamp = response_data["timestamp"]
        dt = datetime.datetime.fromtimestamp(timestamp)
        # 将datetime对象转换为指定格式的字符串
        formatted_datetime = dt.strftime("%Y-%m-%d %H:%M:%S")
        root = tk.Tk()
        root.withdraw()  # 隐藏主窗口
        if error_code == 0:
            # 获取face_num等字段
            face_num = response_data['result']['face_num']
            face_list_text = f"人脸信息列表：\n"
            if face_num > 0:
                face_list = response_data['result']['face_list']
                print(face_list)
                # 遍历列表中的字典
                for item in face_list:
                    face_probability = item['face_probability']
                    age = item["age"]
                    expression = self.determine_expression(item['expression'])
                    face_shape = self.determine_face_shape(item['face_shape'])
                    gender = self.determine_gender(item['gender']['type'])
                    glasses = self.determine_glasses(item['glasses'])
                    emotion = self.determine_emotion(item['emotion'])
                    mask = self.determine_mask(item['mask'])
                    item_text = f"[\n" \
                                f" 人脸置信度（0—1）：{face_probability}\n" \
                                f" 年龄：{age}\n" \
                                f" 表情：{expression}\n" \
                                f" 脸型：{face_shape}\n" \
                                f" 性别：{gender}\n" \
                                f" 是否戴眼镜：{glasses}\n" \
                                f" 情绪：{emotion}\n" \
                                f" 口罩：{mask}\n" \
                                f"]\n"
                    face_list_text = face_list_text + item_text

            messagebox.askokcancel("识别结果", f"识别成功,时间：{formatted_datetime}"
                                               f"\n检测到的图片中的人脸数量：{face_num}"
                                               f"\n{face_list_text}")

        else:
            messagebox.askokcancel("识别结果", "请求失败")
        return response

    def determine_gender(self, gender):
        return '男' if gender == 'male' else '女'

    def determine_expression(self, expression):
        if expression["type"] == "none":
            return "不笑", '置信度:' + str(expression["probability"])
        elif expression["type"] == "smile":
            return "微笑",'置信度:' + str(expression["probability"])
        elif expression["type"] == "laugh":
            return "大笑", '置信度:' + str(expression["probability"])
        else:
            return "无法确定", 0
    def determine_face_shape(self, face_shape):
        if face_shape["type"] == "square":
            return "正方形", '置信度:' + str(face_shape["probability"])
        elif face_shape["type"] == "triangle":
            return "三角形",'置信度:' + str(face_shape["probability"])
        elif face_shape["type"] == "oval":
            return "椭圆", '置信度:' + str(face_shape["probability"])
        elif face_shape["type"] == "heart":
            return "心形", '置信度:' + str(face_shape["probability"])
        elif face_shape["type"] == "round":
            return "圆形", '置信度:' + str(face_shape["probability"])
        else:
            return "无法确定", 0
    def determine_glasses(self, glasses):
        if glasses["type"] == "none":
            return "无眼镜", '置信度:' + str(glasses["probability"])
        elif glasses["type"] == "common":
            return "普通眼镜",'置信度:' + str(glasses["probability"])
        elif glasses["type"] == "sun":
            return "墨镜", '置信度:' + str(glasses["probability"])
        else:
            return "无法确定", 0
    def determine_mask(self, mask):
        if mask["type"] == 0:
            return "没戴口罩", '置信度:' + str(mask["probability"])
        elif mask["type"] == 1:
            return "戴口罩",'置信度:' + str(mask["probability"])
        else:
            return "无法确定", 0
    def determine_emotion(self, emotion):
        if emotion["type"] == "angry":
            return "愤怒", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "disgust":
            return "厌恶",'置信度:' + str(emotion["probability"])
        elif emotion["type"] == "fear":
            return "恐惧", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "happy":
            return "高兴", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "sad":
            return "伤心", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "surprise":
            return "惊讶", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "neutral":
            return "无表情", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "pouty":
            return "噘嘴", '置信度:' + str(emotion["probability"])
        elif emotion["type"] == "grimace":
            return "鬼脸", '置信度:' + str(emotion["probability"])
        else:
            return "无法确定", 0

    def show_file_at_line(self, filename, lineno, col_offset=None):
        editor = self.show_file(filename)
        editor.select_line(lineno, col_offset)

    def update_appearance(self):
        for editor in self.winfo_children():
            editor.update_appearance()

    def update_editor_title(self, editor, title=None):
        if title is None:
            title = editor.get_title()
        try:
            self.tab(editor, text=title)
        except TclError:
            pass

        try:
            self.indicate_modification()
        except Exception:
            logger.exception("Could not update modification indication")

    def indicate_modification(self):
        if not running_on_mac_os():
            return

        atts = self.winfo_toplevel().wm_attributes()
        if "-modified" in atts:
            i = atts.index("-modified")
            mod = atts[i: i + 2]
            rest = atts[:i] + atts[i + 2:]
        else:
            mod = ()
            rest = atts

        for editor in self.get_all_editors():
            if editor.is_modified():
                if mod != ("-modified", 1):
                    self.winfo_toplevel().wm_attributes(*(rest + ("-modified", 1)))
                break
        else:
            if mod == ("-modified", 1):
                self.winfo_toplevel().wm_attributes(*(rest + ("-modified", 0)))

    def _open_file(self, filename):
        editor = Editor(self)
        if editor._load_file(filename):
            self.add(editor, text=editor.get_title())
            return editor
        else:
            editor.destroy()
            return None

    def get_editor(self, filename_or_id, open_when_necessary=False):
        if os.path.isfile(filename_or_id):
            filename_or_id = normpath_with_actual_case(os.path.abspath(filename_or_id))

        for child in self.winfo_children():
            assert isinstance(child, Editor)
            child_identifier = child.get_identifier()
            if child_identifier == filename_or_id:
                return child

        if open_when_necessary:
            return self._open_file(filename_or_id)
        else:
            return None

    def check_allow_closing(self, editor=None):
        if not editor:
            modified_editors = [e for e in self.winfo_children() if e.is_modified()]
        else:
            if not editor.is_modified():
                return True
            else:
                modified_editors = [editor]
        if len(modified_editors) == 0:
            return True

        message = tr("Do you want to save files before closing?")
        if editor:
            message = tr("Do you want to save file before closing?")

        confirm = messagebox.askyesnocancel(
            title=tr("Save On Close"), message=message, default=messagebox.YES, master=self
        )

        if confirm:
            for editor_ in modified_editors:
                assert isinstance(editor_, Editor)
                if editor_.get_filename(True):
                    if not editor_.save_file():
                        return False
                else:
                    return False
            return True

        elif confirm is None:
            return False
        else:
            return True

    def check_for_external_changes(self, event=None):
        if self._checking_external_changes:
            # otherwise the method will be re-entered when focus
            # changes because of a confirmation message box
            return

        self._checking_external_changes = True
        try:
            for editor in self.get_all_editors():
                editor.check_for_external_changes()
        finally:
            self._checking_external_changes = False


# 百度智能云人脸识别
API_KEY = "bo0Tfin2s2Lc3DUVQW3DzOPb"
SECRET_KEY = "DbF0dvfAOW9aLG5Izi1EI3RsVprBwuXp"


def get_access_token():
    """
    使用 AK，SK 生成鉴权签名（Access Token）
    :return: access_token，或是None(如果错误)
    """
    url = "https://aip.baidubce.com/oauth/2.0/token"
    params = {"grant_type": "client_credentials", "client_id": API_KEY, "client_secret": SECRET_KEY}
    return str(requests.post(url, params=params).json().get("access_token"))


def get_file_content_as_base64(path, urlencoded=False):
    """
    获取文件base64编码
    :param path: 文件路径
    :param urlencoded: 是否对结果进行urlencoded
    :return: base64编码信息
    """
    with open(path, "rb") as f:
        content = base64.b64encode(f.read()).decode("utf8")
        if urlencoded:
            content = urllib.parse.quote_plus(content)
    return content


def get_current_breakpoints():
    result = {}

    for editor in get_workbench().get_editor_notebook().get_all_editors():
        filename = editor.get_filename()
        if filename:
            linenos = editor.get_code_view().get_breakpoint_line_numbers()
            if linenos:
                result[filename] = linenos

    return result


def get_saved_current_script_filename(force=True):
    editor = get_workbench().get_editor_notebook().get_current_editor()
    if not editor:
        return None

    filename = editor.get_filename(force)
    if not filename:
        return None

    if editor.is_modified():
        filename = editor.save_file()

    return filename


def get_target_dirname_from_editor_filename(s):
    if is_local_path(s):
        return os.path.dirname(s)
    else:
        return universal_dirname(extract_target_path(s))


def extract_target_path(s):
    assert is_remote_path(s)
    return s[s.find(REMOTE_PATH_MARKER) + len(REMOTE_PATH_MARKER):]


def make_remote_path(target_path):
    return get_runner().get_node_label() + REMOTE_PATH_MARKER + target_path



"""
system specific functions.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/sys.rst.
=======================================

.. module:: sys
   :synopsis: system specific functions

|see_cpython_module| :mod:`python:sys`.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Callable, Final, Literal, NoReturn

from uio import IOBase

class Implementation(tuple[str, tuple[int, int, int], int]):
    name: str
    version: tuple[int, int, int]
    mpy: int

class ModuleType:
    __class__: str
    __name__: str

def exit(retval: object = 0, /) -> NoReturn:
    """
   Terminate current program with a given exit code. Underlyingly, this
   function raise as `SystemExit` exception. If an argument is given, its
   value given as an argument to `SystemExit`.
   """

def atexit(func: Callable[[], None] | None, /) -> Callable[[], None] | None:
    """
   Register *func* to be called upon termination.  *func* must be a callable
   that takes no arguments, or ``None`` to disable the call.  The ``atexit``
   function will return the previous value set by this function, which is
   initially ``None``.
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This function is a MicroPython extension intended to provide similar
      functionality to the :mod:`atexit` module in CPython.
   """

def print_exception(exc: BaseException, file: IOBase[str] = "stdout", /) -> None:
    """
   Print exception with a traceback to a file-like object *file* (or
   `sys.stdout` by default).
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This is simplified version of a function which appears in the
      ``traceback`` module in CPython. Unlike ``traceback.print_exception()``,
      this function takes just exception value instead of exception type,
      exception value, and traceback object; *file* argument should be
      positional; further arguments are not supported. CPython-compatible
      ``traceback`` module can be found in `micropython-lib`.
   
   .. function:: settrace(tracefunc)
   
   Enable tracing of bytecode execution.  For details see the `CPython
   documentaion <https://docs.python.org/3/library/sys.html#sys.settrace>`_.
   
   This function requires a custom MicroPython build as it is typically not
   present in pre-built firmware (due to it affecting performance).  The relevant
   configuration option is *MICROPY_PY_SYS_SETTRACE*.
   """

argv: Final[list[str]] = ...
"""
A mutable list of arguments the current program was started with.
"""

byteorder: Final[Literal["little", "big"]] = ...
"""
The byte order of the system (``"little"`` or ``"big"``).
"""

implementation: Final[Implementation] = ...
"""
Object with information about the current Python implementation. For
   MicroPython, it has following attributes:

   * *name* - string "micropython"
   * *version* - tuple (major, minor, micro), e.g. (1, 7, 0)

   This object is the recommended way to distinguish MicroPython from other
   Python implementations (note that it still may not exist in the very
   minimal ports).

   .. admonition:: Difference to CPython
      :class: attention

      CPython mandates more attributes for this object, but the actual useful
      bare minimum is implemented in MicroPython.
"""

maxsize: Final[int] = ...
"""
Maximum value which a native integer type can hold on the current platform,
   or maximum value representable by MicroPython integer type, if it's smaller
   than platform max value (that is the case for MicroPython ports without
   long int support).

   This attribute is useful for detecting "bitness" of a platform (32-bit vs
   64-bit, etc.). It's recommended to not compare this attribute to some
   value directly, but instead count number of bits in it::

    bits = 0
    v = sys.maxsize
    while v:
        bits += 1
        v >>= 1
    if bits > 32:
        # 64-bit (or more) platform
        ...
    else:
        # 32-bit (or less) platform
        # Note that on 32-bit platform, value of bits may be less than 32
        # (e.g. 31) due to peculiarities described above, so use "> 16",
        # "> 32", "> 64" style of comparisons.
"""

modules: Final[dict[str, ModuleType]] = ...
"""
Dictionary of loaded modules. On some ports, it may not include builtin
   modules.
"""

path: Final[list[str]] = ...
"""
A mutable list of directories to search for imported modules.
"""

platform: Final[str] = ...
"""
The platform that MicroPython is running on. For OS/RTOS ports, this is
   usually an identifier of the OS, e.g. ``"linux"``. For baremetal ports it
   is an identifier of a board, e.g. ``"pyboard"`` for the original MicroPython
   reference board. It thus can be used to distinguish one board from another.
   If you need to check whether your program runs on MicroPython (vs other
   Python implementation), use `sys.implementation` instead.
"""

stderr: Final[IOBase[str]] = ...
"""
Standard error `stream`.
"""

stdin: Final[IOBase[str]] = ...
"""
Standard input `stream`.
"""

stdout: Final[IOBase[str]] = ...
"""
Standard output `stream`.
"""

version: Final[str] = ...
"""
Python language version that this implementation conforms to, as a string.
"""

version_info: Final[tuple[int, int, int]] = ...
"""
Python language version that this implementation conforms to, as a tuple of ints.

    .. admonition:: Difference to CPython
      :class: attention

      Only the first three version numbers (major, minor, micro) are supported and
      they can be referenced only by index, not by name.
"""



"""
See http://microbit-micropython.readthedocs.io/en/latest/speech.html
for details
"""

def translate(words: str) ->None:
    """
    Given English words in the string words, return a string
    containing a best guess at the appropriate phonemes to pronounce.
    The output is generated from this text to phoneme translation table.
    This function should be used to generate a first approximation of phonemes
    that can be further hand-edited to improve accuracy, inflection and emphasis.
    """

def pronounce(phonemes: str, pitch: int=64, speed: int=72,
              mouth: int=128, throat: int=128)->None:
    """
    Pronounce the phonemes in the string phonemes. See below for details of how
    to use phonemes to finely control the output of the speech synthesiser.
    Override the optional pitch, speed, mouth and throat settings to change
    the timbre (quality) of the voice.
    """

def say(words: str, pitch: int=64, speed: int=72,
              mouth: int=128, throat: int=128)->None:
    """
    Say the English words in the string words. The result is semi-accurate
    for English.
    Override the optional pitch, speed, mouth and throat settings to change
    the timbre (quality) of the voice. This is a short-hand equivalent of:
    speech.pronounce(speech.translate(words))
    """

def sing(phonemes: str, pitch: int=64, speed: int=72,
              mouth: int=128, throat: int=128)->None:
    """
    Sing the phonemes contained in the string phonemes. Changing the pitch
    and duration of the note is described below. Override the optional pitch,
    speed, mouth and throat settings to change the timbre (quality) of the voice.
    """



"""functions that an OS normally provides

|see_cpython_module| :mod:`cpython:os`.
"""

from __future__ import annotations

import typing
from typing import Optional, Tuple

def uname() -> _Uname:
    """Returns a named tuple of operating specific and CircuitPython port
    specific information."""
    ...

class _Uname(typing.NamedTuple):
    """The type of values that :py:func:`.uname()` returns"""

    sysname: str
    nodename: str
    release: str
    version: str
    machine: str

def chdir(path: str) -> None:
    """Change current directory."""
    ...

def getcwd() -> str:
    """Get the current directory."""
    ...

def getenv(key: str, default: Optional[str] = None) -> Optional[str]:
    """Get the environment variable value for the given key or return ``default``.

    This may load values from disk so cache the result instead of calling this often."""
    ...

def listdir(dir: str) -> str:
    """With no argument, list the current directory.  Otherwise list the given directory."""
    ...

def mkdir(path: str) -> None:
    """Create a new directory."""
    ...

def remove(path: str) -> None:
    """Remove a file."""
    ...

def rmdir(path: str) -> None:
    """Remove a directory."""
    ...

def rename(old_path: str, new_path: str) -> str:
    """Rename a file."""
    ...

def stat(path: str) -> Tuple[int, int, int, int, int, int, int, int, int, int]:
    """Get the status of a file or directory.

       Returns a tuple with the status of a file or directory in the following order:


       * ``st_mode`` -- File type, regular or directory
       * ``st_ino``  -- Set to 0
       * ``st_dev`` -- Set to 0
       * ``st_nlink`` -- Set to 0
       * ``st_uid`` -- Set to 0
       * ``st_gid`` -- Set to 0
       * ``st_size`` -- Size of the file in bytes
       * ``st_atime`` -- Time of most recent access expressed in seconds
       * ``st_mtime`` -- Time of most recent content modification expressed in seconds.
       * ``st_ctime`` -- Time of most recent content modification expressed in seconds.

    .. note:: On builds without long integers, the number of seconds
       for contemporary dates will not fit in a small integer.
       So the time fields return 946684800,
       which is the number of seconds corresponding to 1999-12-31."""
    ...

def statvfs(path: str) -> Tuple[int, int, int, int, int, int, int, int, int, int]:
    """Get the status of a filesystem.

    Returns a tuple with the filesystem information in the following order:

         * ``f_bsize`` -- file system block size
         * ``f_frsize`` -- fragment size
         * ``f_blocks`` -- size of fs in f_frsize units
         * ``f_bfree`` -- number of free blocks
         * ``f_bavail`` -- number of free blocks for unprivileged users
         * ``f_files`` -- number of inodes
         * ``f_ffree`` -- number of free inodes
         * ``f_favail`` -- number of free inodes for unprivileged users
         * ``f_flag`` -- mount flags
         * ``f_namemax`` -- maximum filename length

    Parameters related to inodes: ``f_files``, ``f_ffree``, ``f_avail``
    and the ``f_flags`` parameter may return ``0`` as they can be unavailable
    in a port-specific implementation."""
    ...

def sync() -> None:
    """Sync all filesystems."""
    ...

def urandom(size: int) -> str:
    """Returns a string of *size* random bytes based on a hardware True Random
    Number Generator. When not available, it will raise a NotImplementedError.

    **Limitations:** Not yet available on nRF. Not available on SAMD21 due to lack of hardware.
    """
    ...

def utime(path: str, times: Tuple[int, int]) -> None:
    """Change the timestamp of a file."""
    ...

sep: str
"""Separator used to delineate path components such as folder and file names."""



"""
binary/ASCII conversions.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/binascii.rst.
===========================================

.. module:: binascii
   :synopsis: binary/ASCII conversions

|see_cpython_module| :mod:`python:binascii`.

This module implements conversions between binary data and various
encodings of it in ASCII form (in both directions).
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

def hexlify(data: bytes, sep: str | bytes = ..., /) -> bytes:
    """
   Convert the bytes in the *data* object to a hexadecimal representation.
   Returns a bytes object.
   
   If the additional argument *sep* is supplied it is used as a separator
   between hexadecimal values.
   """

def unhexlify(data: str | bytes, /) -> bytes:
    """
   Convert hexadecimal data to binary representation. Returns bytes string.
   (i.e. inverse of hexlify)
   """

def a2b_base64(data: str | bytes, /) -> bytes:
    """
   Decode base64-encoded data, ignoring invalid characters in the input.
   Conforms to `RFC 2045 s.6.8 <https://tools.ietf.org/html/rfc2045#section-6.8>`_.
   Returns a bytes object.
   """

def b2a_base64(data: bytes, /) -> bytes:
    """
   Encode binary data in base64 format, as in `RFC 3548
   <https://tools.ietf.org/html/rfc3548.html>`_. Returns the encoded data
   followed by a newline character, as a bytes object.
   """



"""Multicast Domain Name Service

The `mdns` module provides basic support for multicast domain name services.
Basic use provides hostname resolution under the .local TLD. This module
also supports DNS Service Discovery that allows for discovering other hosts
that provide a desired service."""

from __future__ import annotations

import ipaddress
from typing import Optional, Tuple

import wifi

class RemoteService:
    """Encapsulates information about a remote service that was found during a search. This
    object may only be created by a `mdns.Server`. It has no user-visible constructor.
    """

    def __init__(self) -> None:
        """Cannot be instantiated directly. Use `mdns.Server.find`."""
        ...
    hostname: str
    """The hostname of the device (read-only),."""
    instance_name: str
    """The human readable instance name for the service. (read-only)"""
    service_type: str
    """The service type string such as ``_http``. (read-only)"""
    protocol: str
    """The protocol string such as ``_tcp``. (read-only)"""
    port: int
    """Port number used for the service. (read-only)"""
    ipv4_address: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the remote service. None if no A records are found."""
    def __del__(self) -> None:
        """Deletes the RemoteService object."""
        ...

class Server:
    """The MDNS Server responds to queries for this device's information and allows for querying
    other devices."""

    def __init__(self, network_interface: wifi.Radio) -> None:
        """
        Constructs or returns the mdns.Server for the given network_interface. (CircuitPython
        may already be using it.) Only native interfaces are currently supported.
        """
        ...
    def deinit(self) -> None:
        """Stops the server"""
        ...
    hostname: str
    """Hostname resolvable as ``<hostname>.local`` in addition to ``circuitpython.local``. Make
       sure this is unique across all devices on the network. It defaults to ``cpy-######``
       where ``######`` is the hex digits of the last three bytes of the mac address."""
    instance_name: str
    """Human readable name to describe the device."""
    def find(
        self, service_type: str, protocol: str, *, timeout: float = 1
    ) -> Tuple[RemoteService]:
        """Find all locally available remote services with the given service type and protocol.

        This doesn't allow for direct hostname lookup. To do that, use
        `socketpool.SocketPool.getaddrinfo()`.

        :param str service_type: The service type such as "_http"
        :param str protocol: The service protocol such as "_tcp"
        :param float/int timeout: Time to wait for responses"""
        ...
    def advertise_service(self, *, service_type: str, protocol: str, port: int) -> None:
        """Respond to queries for the given service with the given port.

        ``service_type`` and ``protocol`` can only occur on one port. Any call after the first
        will update the entry's port.

        :param str service_type: The service type such as "_http"
        :param str protocol: The service protocol such as "_tcp"
        :param int port: The port used by the service"""
        ...



"""
cryptographic ciphers.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/cryptolib.rst.
=========================================

.. module:: cryptolib
   :synopsis: cryptographic ciphers
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import overload

from uio import AnyReadableBuf, AnyWritableBuf

# noinspection PyPep8Naming
class aes:
    """
   .. class:: aes
   """

    @overload
    def __init__(self, key: AnyReadableBuf, mode: int, /):
        """
           Initialize cipher object, suitable for encryption/decryption. Note:
           after initialization, cipher object can be use only either for
           encryption or decryption. Running decrypt() operation after encrypt()
           or vice versa is not supported.
           
           Parameters are:
           
               * *key* is an encryption/decryption key (bytes-like).
               * *mode* is:
           
                   * ``1`` (or ``cryptolib.MODE_ECB`` if it exists) for Electronic Code Book (ECB).
                   * ``2`` (or ``cryptolib.MODE_CBC`` if it exists) for Cipher Block Chaining (CBC).
                   * ``6`` (or ``cryptolib.MODE_CTR`` if it exists) for Counter mode (CTR).
           
               * *IV* is an initialization vector for CBC mode.
               * For Counter mode, *IV* is the initial value for the counter.
      """
    @overload
    def __init__(self, key: AnyReadableBuf, mode: int, IV: AnyReadableBuf, /):
        """
           Initialize cipher object, suitable for encryption/decryption. Note:
           after initialization, cipher object can be use only either for
           encryption or decryption. Running decrypt() operation after encrypt()
           or vice versa is not supported.
           
           Parameters are:
           
               * *key* is an encryption/decryption key (bytes-like).
               * *mode* is:
           
                   * ``1`` (or ``cryptolib.MODE_ECB`` if it exists) for Electronic Code Book (ECB).
                   * ``2`` (or ``cryptolib.MODE_CBC`` if it exists) for Cipher Block Chaining (CBC).
                   * ``6`` (or ``cryptolib.MODE_CTR`` if it exists) for Counter mode (CTR).
           
               * *IV* is an initialization vector for CBC mode.
               * For Counter mode, *IV* is the initial value for the counter.
      """
    @overload
    def encrypt(self, in_buf: AnyReadableBuf, /) -> bytes:
        """
           Encrypt *in_buf*. If no *out_buf* is given result is returned as a
           newly allocated `bytes` object. Otherwise, result is written into
           mutable buffer *out_buf*. *in_buf* and *out_buf* can also refer
           to the same mutable buffer, in which case data is encrypted in-place.
      """
    @overload
    def encrypt(self, in_buf: AnyReadableBuf, out_buf: AnyWritableBuf, /) -> None:
        """
           Encrypt *in_buf*. If no *out_buf* is given result is returned as a
           newly allocated `bytes` object. Otherwise, result is written into
           mutable buffer *out_buf*. *in_buf* and *out_buf* can also refer
           to the same mutable buffer, in which case data is encrypted in-place.
      """
    @overload
    def decrypt(self, in_buf: AnyReadableBuf, /) -> bytes:
        """
           Like `encrypt()`, but for decryption.
      """
    @overload
    def decrypt(self, in_buf: AnyReadableBuf, out_buf: AnyWritableBuf, /) -> None:
        """
           Like `encrypt()`, but for decryption.
      """



"""
Publishes the classes imported from the submodules.
"""

from .app import App
from .button import Button
from .colorsensor import ColorSensor
from .distancesensor import DistanceSensor
from .forcesensor import ForceSensor
from .lightmatrix import LightMatrix
from .motionsensor import MotionSensor
from .motor import Motor
from .motorpair import MotorPair
from .speaker import Speaker
from .statuslight import StatusLight


"""
functions related to the ESP8266 and ESP32.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/esp.rst.
=========================================================

.. module:: esp
    :synopsis: functions related to the ESP8266 and ESP32

The ``esp`` module contains specific functions related to both the ESP8266 and 
ESP32 modules.  Some functions are only available on one or the other of these
ports.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final, overload

from uio import AnyWritableBuf, AnyReadableBuf

SLEEP_NONE: Final[int] = ...
"""All functions enabled."""

SLEEP_MODEM: Final[int] = ...
"""Modem sleep, shuts down the WiFi Modem circuit."""

SLEEP_LIGHT: Final[int] = ...
"""Light sleep, shuts down the WiFi Modem circuit."""

# noinspection PyShadowingNames
@overload
def sleep_type(sleep_type: int, /) -> None:
    """
    **Note**: ESP8266 only
    
    Get or set the sleep type.
    
    If the *sleep_type* parameter is provided, sets the sleep type to its
    value. If the function is called without parameters, returns the current
    sleep type.
    
    The possible sleep types are defined as constants:
    
        * ``SLEEP_NONE`` -- all functions enabled,
        * ``SLEEP_MODEM`` -- modem sleep, shuts down the WiFi Modem circuit.
        * ``SLEEP_LIGHT`` -- light sleep, shuts down the WiFi Modem circuit
          and suspends the processor periodically.
    
    The system enters the set sleep mode automatically when possible.
   """

# noinspection PyShadowingNames
@overload
def sleep_type() -> int:
    """
    **Note**: ESP8266 only
    
    Get or set the sleep type.
    
    If the *sleep_type* parameter is provided, sets the sleep type to its
    value. If the function is called without parameters, returns the current
    sleep type.
    
    The possible sleep types are defined as constants:
    
        * ``SLEEP_NONE`` -- all functions enabled,
        * ``SLEEP_MODEM`` -- modem sleep, shuts down the WiFi Modem circuit.
        * ``SLEEP_LIGHT`` -- light sleep, shuts down the WiFi Modem circuit
          and suspends the processor periodically.
    
    The system enters the set sleep mode automatically when possible.
   """

def deepsleep(time_us: int = 0, /) -> None:
    """
    **Note**: ESP8266 only - use `machine.deepsleep()` on ESP32
    
    Enter deep sleep.
    
    The whole module powers down, except for the RTC clock circuit, which can
    be used to restart the module after the specified time if the pin 16 is
    connected to the reset pin. Otherwise the module will sleep until manually
    reset.
   """

def flash_id() -> int:
    """
    **Note**: ESP8266 only
    
    Read the device ID of the flash memory.
   """

def flash_size() -> int:
    """
    Read the total size of the flash memory.
   """

def flash_user_start() -> int:
    """
    Read the memory offset at which the user flash space begins.
   """

@overload
def flash_read(byte_offset: int, length_or_buffer: int, /) -> bytes:
    """
   Reads bytes from the flash memory starting at the given byte offset.
   If length is specified: reads the given length of bytes and returns them as ``bytes``.
   If a buffer is given: reads the buf length of bytes and writes them into the buffer.
   Note: esp32 doesn't support passing a length, just a buffer.
   """

@overload
def flash_read(byte_offset: int, length_or_buffer: AnyWritableBuf, /) -> None:
    """
   Reads bytes from the flash memory starting at the given byte offset.
   If length is specified: reads the given length of bytes and returns them as ``bytes``.
   If a buffer is given: reads the buf length of bytes and writes them into the buffer.
   Note: esp32 doesn't support passing a length, just a buffer.
   """

def flash_write(byte_offset: int, bytes: AnyReadableBuf, /) -> None:
    """
   Writes given bytes buffer to the flash memory starting at the given byte offset.
   """

def flash_erase(sector_no: int, /) -> None:
    """
   Erases the given *sector* of flash memory.
   """

@overload
def set_native_code_location(start: None, length: None, /) -> None:
    """
    **Note**: ESP8266 only
    
    Set the location that native code will be placed for execution after it is
    compiled.  Native code is emitted when the ``@micropython.native``,
    ``@micropython.viper`` and ``@micropython.asm_xtensa`` decorators are applied
    to a function.  The ESP8266 must execute code from either iRAM or the lower
    1MByte of flash (which is memory mapped), and this function controls the
    location.
    
    If *start* and *length* are both ``None`` then the native code location is
    set to the unused portion of memory at the end of the iRAM1 region.  The
    size of this unused portion depends on the firmware and is typically quite
    small (around 500 bytes), and is enough to store a few very small
    functions.  The advantage of using this iRAM1 region is that it does not
    get worn out by writing to it.
    
    If neither *start* nor *length* are ``None`` then they should be integers.
    *start* should specify the byte offset from the beginning of the flash at
    which native code should be stored.  *length* specifies how many bytes of
    flash from *start* can be used to store native code.  *start* and *length*
    should be multiples of the sector size (being 4096 bytes).  The flash will
    be automatically erased before writing to it so be sure to use a region of
    flash that is not otherwise used, for example by the firmware or the
    filesystem.
    
    When using the flash to store native code *start+length* must be less
    than or equal to 1MByte.  Note that the flash can be worn out if repeated
    erasures (and writes) are made so use this feature sparingly.
    In particular, native code needs to be recompiled and rewritten to flash
    on each boot (including wake from deepsleep).
    
    In both cases above, using iRAM1 or flash, if there is no more room left
    in the specified region then the use of a native decorator on a function
    will lead to `MemoryError` exception being raised during compilation of
    that function.
   """

@overload
def set_native_code_location(start: int, length: int, /) -> None:
    """
    **Note**: ESP8266 only
    
    Set the location that native code will be placed for execution after it is
    compiled.  Native code is emitted when the ``@micropython.native``,
    ``@micropython.viper`` and ``@micropython.asm_xtensa`` decorators are applied
    to a function.  The ESP8266 must execute code from either iRAM or the lower
    1MByte of flash (which is memory mapped), and this function controls the
    location.
    
    If *start* and *length* are both ``None`` then the native code location is
    set to the unused portion of memory at the end of the iRAM1 region.  The
    size of this unused portion depends on the firmware and is typically quite
    small (around 500 bytes), and is enough to store a few very small
    functions.  The advantage of using this iRAM1 region is that it does not
    get worn out by writing to it.
    
    If neither *start* nor *length* are ``None`` then they should be integers.
    *start* should specify the byte offset from the beginning of the flash at
    which native code should be stored.  *length* specifies how many bytes of
    flash from *start* can be used to store native code.  *start* and *length*
    should be multiples of the sector size (being 4096 bytes).  The flash will
    be automatically erased before writing to it so be sure to use a region of
    flash that is not otherwise used, for example by the firmware or the
    filesystem.
    
    When using the flash to store native code *start+length* must be less
    than or equal to 1MByte.  Note that the flash can be worn out if repeated
    erasures (and writes) are made so use this feature sparingly.
    In particular, native code needs to be recompiled and rewritten to flash
    on each boot (including wake from deepsleep).
    
    In both cases above, using iRAM1 or flash, if there is no more room left
    in the specified region then the use of a native decorator on a function
    will lead to `MemoryError` exception being raised during compilation of
    that function.
   """



"""Analog Buffered IO Hardware Support

The `analogbufio` module contains classes to provide access to analog-to-digital
conversion and digital-to-analog (DAC) for multiple value transfer.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

TODO: For the essentials of `analogbufio`, see the `CircuitPython Essentials
Learn guide <https://learn.adafruit.com/circuitpython-essentials/circuitpython-analogbufio>`_

TODO: For more information on using `analogbufio`, see `this additional Learn guide
<https://learn.adafruit.com/circuitpython-advanced-analog-inputs-and-outputs>`_
"""

from __future__ import annotations

import microcontroller
from circuitpython_typing import WriteableBuffer

class BufferedIn:
    """Capture multiple analog voltage levels to the supplied buffer

    Usage::

        import board
        import analogbufio
        import array

        length = 1000
        mybuffer = array.array("H", [0x0000] * length)
        rate = 500000
        adcbuf = analogbufio.BufferedIn(board.GP26, sample_rate=rate)
        adcbuf.readinto(mybuffer)
        adcbuf.deinit()
        for i in range(length):
            print(i, mybuffer[i])

        (TODO) The reference voltage varies by platform so use
        ``reference_voltage`` to read the configured setting.
        (TODO) Provide mechanism to read CPU Temperature."""

    def __init__(self, pin: microcontroller.Pin, *, sample_rate: int) -> None:
        """Create a `BufferedIn` on the given pin and given sample rate.

        :param ~microcontroller.Pin pin: the pin to read from
        :param ~int sample_rate: rate: sampling frequency, in samples per second"""
        ...
    def deinit(self) -> None:
        """Shut down the `BufferedIn` and release the pin for other use."""
        ...
    def __enter__(self) -> BufferedIn:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def readinto(self, buffer: WriteableBuffer) -> int:
        """Fills the provided buffer with ADC voltage values.

        ADC values will be read into the given buffer at the supplied sample_rate.
        Depending on the buffer typecode, 'B', 'H', samples are 8-bit byte-arrays or
        16-bit half-words and are always unsigned.
        The ADC most significant bits of the ADC are kept. (See
        https://docs.circuitpython.org/en/latest/docs/library/array.html)

        :param ~circuitpython_typing.WriteableBuffer buffer: buffer: A buffer for samples
        """
        ...



"""Support for frequency based protocols

.. warning:: This module is not available in SAMD21 builds. See the
  :ref:`module-support-matrix` for more info.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import time
  import frequencyio
  import board

  frequency = frequencyio.FrequencyIn(board.D11)
  frequency.capture_period = 15
  time.sleep(0.1)

This example will initialize the the device, set
:py:data:`~frequencyio.FrequencyIn.capture_period`, and then sleep 0.1 seconds.
CircuitPython will automatically turn off FrequencyIn capture when it resets all
hardware after program completion. Use ``deinit()`` or a ``with`` statement
to do it yourself."""

from __future__ import annotations

import microcontroller

class FrequencyIn:
    """Read a frequency signal

    FrequencyIn is used to measure the frequency, in hertz, of a digital signal
    on an incoming pin. Accuracy has shown to be within 10%, if not better. It
    is recommended to utilize an average of multiple samples to smooth out readings.

    Frequencies below 1KHz are not currently detectable.

    FrequencyIn will not determine pulse width (use ``PulseIn``)."""

    def __init__(self, pin: microcontroller.Pin, capture_period: int = 10) -> None:
        """Create a FrequencyIn object associated with the given pin.

        :param ~microcontroller.Pin pin: Pin to read frequency from.
        :param int capture_period: Keyword argument to set the measurement period, in
                                   milliseconds. Default is 10ms; range is 1ms - 500ms.

        Read the incoming frequency from a pin::

          import frequencyio
          import board

          frequency = frequencyio.FrequencyIn(board.D11)

          # Loop while printing the detected frequency
          while True:
              print(frequency.value)

              # Optional clear() will reset the value
              # to zero. Without this, if the incoming
              # signal stops, the last reading will remain
              # as the value.
              frequency.clear()"""
        ...
    def deinit(self) -> None:
        """Deinitialises the FrequencyIn and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> FrequencyIn:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def pause(self) -> None:
        """Pause frequency capture."""
        ...
    def resume(self) -> None:
        """Resumes frequency capture."""
        ...
    def clear(self) -> None:
        """Clears the last detected frequency capture value."""
        ...
    capture_period: int
    """The capture measurement period. Lower incoming frequencies will be measured
    more accurately with longer capture periods. Higher frequencies are more
    accurate with shorter capture periods.

    .. note:: When setting a new ``capture_period``, all previous capture information is
              cleared with a call to ``clear()``."""
    def __get__(self, index: int) -> int:
        """Returns the value of the last frequency captured."""
        ...



# -*- coding: utf-8 -*-
from logging import getLogger
from tkinter import ttk

from turcar import get_runner, get_workbench
from turcar.common import InlineCommand
from turcar.languages import tr
from turcar.memory import VariablesFrame

logger = getLogger(__name__)


class VariablesView(VariablesFrame):
    # TODO: Indicate invalid state when program or debug command is running more than a second
    def __init__(self, master):
        super().__init__(master)

        ttk.Style().configure("Centered.TButton", justify="center")
        self.back_button = ttk.Button(
            self.tree,
            style="Centered.TButton",
            text=tr("Back to\ncurrent frame"),
            command=self._handle_back_button,
            width=15,
        )

        get_workbench().bind("BackendRestart", self._handle_backend_restart, True)
        get_workbench().bind("ToplevelResponse", self._handle_toplevel_response, True)
        # get_workbench().bind("DebuggerResponse", self._debugger_response, True)
        get_workbench().bind("get_frame_info_response", self._handle_frame_info_event, True)
        get_workbench().bind("get_globals_response", self._handle_get_globals_response, True)

        # records last info from progress messages
        self._last_active_info = None

    def _update_back_button(self, visible):
        if visible:
            assert self._last_active_info is not None
            self.back_button.configure(text=tr("Back to\n%s") % self._last_active_info[-1])
            self.back_button.place(relx=1, x=-5, y=5, anchor="ne")
        else:
            self.back_button.place_forget()

    def _handle_back_button(self):
        assert self._last_active_info is not None
        if len(self._last_active_info) == 2:
            self.show_globals(*self._last_active_info)
        else:
            assert len(self._last_active_info) == 4
            self.show_frame_variables(*self._last_active_info)

    def _handle_backend_restart(self, event):
        self._clear_tree()

    def _handle_get_globals_response(self, event):
        if "error" in event:
            self._handle_error_response(event["error"])
        elif "globals" not in event:
            self._handle_error_response(str(event))
        else:
            self.show_globals(event["globals"], event["module_name"])

    def _handle_error_response(self, error_msg):
        self._clear_tree()
        logger.error("Error querying globals: %s", error_msg)
        self.show_error("Could not query global variables: " + str(error_msg))

    def _handle_toplevel_response(self, event):
        if "globals" in event:
            self.show_globals(event["globals"], "__main__")
        else:
            # MicroPython
            get_runner().send_command(InlineCommand("get_globals", module_name="__main__"))

    def show_globals(self, globals_, module_name, is_active=True):
        self.clear_error()
        # TODO: update only if something has changed
        self.update_variables(globals_)

        if self.containing_notebook is not None:
            if module_name == "__main__":
                self.containing_notebook.tab(self, text=get_default_tab_text())
            else:
                self.containing_notebook.tab(self, text=get_default_tab_text() + " (%s)" % module_name)

        if is_active:
            self._last_active_info = (globals_, module_name)

        self._update_back_button(not is_active)

    def show_frame_variables(self, locals_, globals_, freevars, frame_name, is_active=True):
        # TODO: update only if something has changed
        actual_locals = {}
        nonlocals = {}
        for name in locals_:
            if name in freevars:
                nonlocals[name] = locals_[name]
            else:
                actual_locals[name] = locals_[name]

        groups = [("LOCALS", actual_locals), ("GLOBALS", globals_)]
        if nonlocals:
            groups.insert(1, ("NONLOCALS", nonlocals))

        self.update_variables(groups)
        if self.containing_notebook is not None:
            self.containing_notebook.tab(self, text=get_default_tab_text() + " (%s)" % frame_name)

        if is_active:
            self._last_active_info = (locals_, globals_, freevars, frame_name)

        self._update_back_button(not is_active)

    def _handle_frame_info_event(self, frame_info):
        if frame_info.get("error"):
            "probably non-existent frame"
            return
        else:
            is_active = frame_info[
                "location"
            ] == "stack" or (  # same __main__ globals can be in different frames
                frame_info["code_name"] == "<module>"
                and frame_info["module_name"] == "__main__"
                and self._last_active_info[-1] == "__main__"
                and self._last_active_info[0] == frame_info["globals"]
            )

            if frame_info["code_name"] == "<module>":
                self.show_globals(frame_info["globals"], frame_info["module_name"], is_active)
            else:
                self.show_frame_variables(
                    frame_info["locals"],
                    frame_info["globals"],
                    frame_info["freevars"],
                    frame_info["code_name"],
                    is_active,
                )


def get_default_tab_text() -> str:
    return tr("Variables")


def load_plugin() -> None:
    get_workbench().add_view(
        VariablesView, get_default_tab_text(), "ne", default_position_key="AAA"
    )



import os.path
import subprocess
import sys

from turcar import get_workbench
from turcar.languages import tr
from turcar.ui_utils import select_sequence


def print_current_script():
    editor = _get_current_editor()
    assert editor is not None

    template_fn = os.path.join(os.path.dirname(__file__), "template.html")
    with open(template_fn, encoding="utf-8") as f:
        template_html = f.read()

    script_html = _export_text_as_html(editor.get_text_widget())
    title_html = escape_html(editor.get_title())
    line_count = last_line = int(float(editor.get_text_widget().index("end-1c")))
    if line_count < 10:
        gutter_ems = 1
    elif line_count < 100:
        gutter_ems = 1.5
    elif line_count < 1000:
        gutter_ems = 2
    else:
        gutter_ems = 3

    full_html = (
        template_html.replace("%title%", title_html)
        .replace("%script%", script_html)
        .replace("%gutter_ems%", str(gutter_ems))
    )

    import tempfile

    temp_handle, temp_fn = tempfile.mkstemp(
        suffix=".html", prefix="turcar_", dir=get_workbench().get_temp_dir()
    )
    with os.fdopen(temp_handle, "w", encoding="utf-8") as f:
        f.write(full_html)

    if sys.platform == "darwin":
        subprocess.Popen(["open", temp_fn])
    else:
        import webbrowser

        webbrowser.open(temp_fn)


def can_print_current_script():
    return _get_current_editor() is not None


def _export_text_as_html(text):
    last_line = int(float(text.index("end-1c")))
    result = ""
    for i in range(1, last_line + 1):
        result += "<code>" + _export_line_as_html(text, i) + "</code>\n"
    return result


def _export_line_as_html(text, lineno):
    s = text.get("%d.0" % lineno, "%d.0 lineend" % lineno).strip("\r\n")

    parts = []
    for i in range(len(s)):
        tag_names = text.tag_names("%d.%d" % (lineno, i))
        if not parts or parts[-1][1] != tag_names:
            parts.append([s[i], tag_names])
        else:
            parts[-1][0] += s[i]

    # print(lineno, parts)
    result = ""
    for s, tags in parts:
        if tags:
            result += "<span class='%s'>%s</span>" % (" ".join(tags), escape_html(s))
        else:
            result += escape_html(s)

    return result


def escape_html(s):
    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def _get_current_editor():
    return get_workbench().get_editor_notebook().get_current_editor()


def load_plugin():
    get_workbench().add_command(
        "printcurrent",
        "file",
        tr("Print..."),
        print_current_script,
        can_print_current_script,
        default_sequence=select_sequence("<Control-p>", "<Command-p>"),
        extra_sequences=["<Control-Greek_pi>"],
        group=11,
    )



from urllib.request import urlopen

import pymysql
import threading
import json
import time
import hashlib
import base64
import requests
import urllib

# 摄像头配置
url = 'http://192.168.0.200:80/'
img = ''
bts = bytes()
is_exit = False
CAMERA_BUFFER_SIZE = 4096

# MySQL 配置
mysql_host = '180.76.235.69'
mysql_user = 'root'
mysql_password = ' a!Y)iiBTF3.t'
mysql_database = 'face-recognition'
charset = 'utf8mb4'

# 百度智能云bos
bos_host = "https://bj.bcebos.com"
access_key_id = "1ee18128a39d4b69881c222bce50e8f6"
secret_access_key = "ee6e8cfb4b5b45ee8136f7ce43eb32a3"
bucket_name = 'yinhang-prod'
# 百度智能云人脸识别
API_KEY = "bo0Tfin2s2Lc3DUVQW3DzOPb"
SECRET_KEY = "DbF0dvfAOW9aLG5Izi1EI3RsVprBwuXp"


def save_image_to_bos3(data, content_length, content_md5):
    pass
    # # 创建BceClientConfiguration
    # config = BceClientConfiguration(credentials=BceCredentials(access_key_id, secret_access_key), endpoint=bos_host)
    # # 创建 BOS 客户端配置
    # client = BosClient(config)
    # # 指定存储路径
    # file_key = "yubo/" + str(int(time.time())) + ".jpg"
    # # print(f"bucket_name:{bucket_name}\ncontent_length:{content_length}\naccess_key_id:{access_key_id}\ndata:{data}\ncontent_md5:{content_md5}")
    # client.put_object(bucket_name, file_key, data, content_length, content_md5)
    #
    # # 返回文件在 BOS 中的完整路径
    # return f"https://{bucket_name}.bj.bcebos.com/{file_key}"


def save_image_path_to_mysql(image_path):
    db_host = mysql_host
    db_user = mysql_user
    db_password = mysql_password
    db_name = mysql_database

    conn = pymysql.connect(host=db_host, user=db_user, password=db_password, database=db_name)
    cursor = conn.cursor()
    sql = "INSERT INTO t_picture(picture_path) VALUES (%s)"
    print(sql)
    cursor.execute(sql, image_path)
    conn.commit()

    cursor.close()
    conn.close()


# 人脸识别
def faceReco(image):
    url = "https://aip.baidubce.com/rest/2.0/face/v3/detect?access_token=" + get_access_token()

    # image 可以通过 get_file_content_as_base64("C:\fakepath\多人二维码.jpg",False) 方法获取
    payload = json.dumps({
        "image": image,
        "image_type": "BASE64"
    })
    headers = {
        'Content-Type': 'application/json'
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    print("照片识别结果：" + response.text)
    return response


def get_file_content_as_base64(path, urlencoded=False):
    """
    获取文件base64编码
    :param path: 文件路径
    :param urlencoded: 是否对结果进行urlencoded
    :return: base64编码信息
    """
    with open(path, "rb") as f:
        content = base64.b64encode(f.read()).decode("utf8")
        if urlencoded:
            content = urllib.parse.quote_plus(content)
    return content


def get_access_token():
    """
    使用 AK，SK 生成鉴权签名（Access Token）
    :return: access_token，或是None(如果错误)
    """
    url = "https://aip.baidubce.com/oauth/2.0/token"
    params = {"grant_type": "client_credentials", "client_id": API_KEY, "client_secret": SECRET_KEY}
    return str(requests.post(url, params=params).json().get("access_token"))


def read_ip_camera():
    global url
    url = 'http://192.168.0.200:80/'
    global img
    global bts
    global is_exit

    count = 0
    last_capture_time = time.time()  # 记录上一次抓取照片的时间
    stream = urlopen(url)
    ts = b''
    img = None
    is_exit = False
    while True:
        if is_exit:
            break
        try:
            # 抓取照片的条件：满足一分钟时间间隔
            if time.time() - last_capture_time >= 3.0:
                bts += stream.read(CAMERA_BUFFER_SIZE)
                # JPEG 图像的起始和结束标记
                jpghead = bts.find(b'\xff\xd8')
                jpgend = bts.find(b'\xff\xd9')
                # 如果找到起始和结束标记，则提取 JPEG 图像的字节数据，
                # 并使用 cv2.imdecode() 函数解码为图像。
                # np.frombuffer(jpg, dtype=np.uint8) 将字节数据转换为 NumPy 数组。
                if jpghead > -1 and jpgend > -1:
                    print(f"图片数量{count}")
                    count += 1
                    jpg = bts[jpghead:jpgend + 2]
                    bts = bts[jpgend + 2:]
                    # 计算哈希值
                    md5_hash = hashlib.md5(jpg).digest()
                    content_md5 = base64.standard_b64encode(md5_hash)
                    # 将哈希值进行 base64 编码
                    content_utf8 = base64.b64encode(jpg).decode("utf8")
                    # content_utf8 = base64.standard_b64encode(md5_hash).decode('utf-8')
                    # 人脸识别，如果有人脸才保存照片
                    # face_num = faceReco(content_utf8).josn().get("result", {}).get("face_num")
                    response = faceReco(content_utf8)
                    if response is not None:
                        error_code = response.json().get("error_code")
                        print(error_code)
                        if error_code == 0:
                            image_path = save_image_to_bos3(jpg, len(jpg), content_md5)
                            print(image_path)
                            save_image_path_to_mysql(image_path)
                last_capture_time = time.time()  # 更新上一次抓取照片的时间
        except Exception as e:
            print("Error:" + str(e))
            bts = b''
            stream = urlopen(url)
            continue


if __name__ == '__main__':
    # print(get_access_token)
    t = threading.Thread(target=read_ip_camera)
    t.daemon = True
    t.start()

    count = 0
    while True:
        time.sleep(1)
        # count += 1
        # print(f"Processed {count} frames")
        # if count >= 10:
        #     is_exit = True
        #     break


"""Hardware accelerated external bus access

The I2CDevice and SPIDevice helper classes make managing transaction state on a bus easy.
For example, they manage locking the bus to prevent other concurrent access. For SPI
devices, it manages the chip select and protocol changes such as mode. For I2C, it
manages the device address."""

from __future__ import annotations



from turcar.languages import tr
from turcar.plugins.circuitpython.cirpy_front import CircuitPythonConfigPage, CircuitPythonProxy
from turcar.plugins.micropython import add_micropython_backend


def load_plugin():
    add_micropython_backend(
        "CircuitPython",
        CircuitPythonProxy,
        tr("CircuitPython (generic)"),
        CircuitPythonConfigPage,
        sort_key="50",
    )



"""
The neopixel module lets you use Neopixel (WS2812) individually
addressable RGB LED strips with the Microbit.

Note
From our tests, the Microbit Neopixel module can drive up to around 256
Neopixels. Anything above that and you may experience weird bugs and issues.
NeoPixels are fun strips of multi-coloured programmable LEDs.
This module contains everything to plug them into a micro:bit
and create funky displays, art and games

Warning
Do not use the 3v connector on the Microbit to power any more than
8 Neopixels at a time.
If you wish to use more than 8 Neopixels, you must use a separate
3v-5v power supply for the Neopixel power pin.

Operations
Writing the colour doesn’t update the display (use show() for that).
np[0] = (255, 0, 128)  # first element
np[-1] = (0, 255, 0)  # last element
np.show()  # only now will the updated value be shown

To read the colour of a specific pixel just reference it.
print(np[0])

Using Neopixels
Interact with Neopixels as if they were a list of tuples.
Each tuple represents the RGB (red, green and blue) mix of colours
for a specific pixel. The RGB values can range between 0 to 255.
For example, initialise a strip of 8 neopixels on a strip connected
to pin0 like this:
import neopixel
np = neopixel.NeoPixel(pin0, 8)
Set pixels by indexing them (like with a Python list). For instance,
to set the first pixel to full brightness red, you would use:
np[0] = (255, 0, 0)
Or the final pixel to purple:
np[-1] = (255, 0, 255)
Get the current colour value of a pixel by indexing it. For example,
to print the first pixel’s RGB value use:
print(np[0])
Finally, to push the new colour data to your Neopixel strip, use the .show() function:
np.show()
If nothing is happening, it’s probably because you’ve forgotten this final step..!
Note
If you’re not seeing anything change on your Neopixel strip,
make sure you have show() at least somewhere otherwise your updates won’t be shown.
"""
from typing import Tuple, Union

from .microbit import MicroBitAnalogDigitalPin

def NeoPixel(pin: MicroBitAnalogDigitalPin, n: int, bpp=...)-> None:
    """
    Initialise a new strip of n number of neopixel LEDs controlled via pin pin.
    The V2 micro:bit can also support RGBW neopixels, so a third argument can be passed to NeoPixel
    to indicate the number of bytes per pixel (bpp). For RGBW, this is is 4 rather than the default
    of 3 for RGB and GRB.

    Each pixel is addressed by a position (starting from 0). Neopixels are given RGB
    (red, green, blue) / RGBW (red, green, blue, white) V2 values between 0-255 as a tuple.
    For example, in RGB, (255,255,255) is white. In RGBW, (255,255,255,0) or (0,0,0,255) is white.
    """

def clear()->None:
    """
    Clear all the pixels.
    """

def show()->None:
    """
    Show the pixels. Must be called for any updates to become visible.
    """

def write()->None:
    """
    Available in V2, same as `show`.
    """

def fill(colour: Union[Tuple[int, int, int], Tuple[int, int, int, int]]) -> None:
    """
    Available in V2. Colour all pixels a given RGB/RGBW value. The colour argument should be a tuple of the same
    length as the number of bytes per pixel (bpp). For example fill((0,0,255)). Use in conjunction
    with show() to update the Neopixels.
    """



"""
zlib decompression.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/zlib.rst.
=================================

.. module:: zlib
   :synopsis: zlib decompression

|see_cpython_module| :mod:`python:zlib`.

This module allows to decompress binary data compressed with
`DEFLATE algorithm <https://en.wikipedia.org/wiki/DEFLATE>`_
(commonly used in zlib library and gzip archiver). Compression
is not yet implemented.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from uio import IOBase

def decompress(data: bytes, wbits: int = 0, bufsize: int = 0, /) -> bytes:
    """
   Return decompressed *data* as bytes. *wbits* is DEFLATE dictionary window
   size used during compression (8-15, the dictionary size is power of 2 of
   that value). Additionally, if value is positive, *data* is assumed to be
   zlib stream (with zlib header). Otherwise, if it's negative, it's assumed
   to be raw DEFLATE stream. *bufsize* parameter is for compatibility with
   CPython and is ignored.
   """

class DecompIO(IOBase[bytes]):
    """
   Steam wrapper that decompresses a given stream containing zlib compressed data.
   """

    def __init__(self, stream: IOBase[bytes], wbits: int = 0, /):
        """
      Create a `stream` wrapper which allows transparent decompression of
      compressed data in another *stream*. This allows to process compressed
      streams with data larger than available heap size. In addition to
      values described in :func:`decompress`, *wbits* may take values
      24..31 (16 + 8..15), meaning that input stream has gzip header.
      
      .. admonition:: Difference to CPython
         :class: attention
      
         This class is MicroPython extension. It's included on provisional
         basis and may be changed considerably or removed in later versions.
      """



from turcar.plugins.pip_gui import _EXTRA_MARKER_RE


def test_extra_marker_re() -> None:
    for text in [
        "extra == 'dev'",
        r"extra == 'dev \' with quote'",
        "  extra  ==  'dev'  ",
        'extra == "dev"',
        r'extra == "dev \"with quote"',
    ]:
        assert _EXTRA_MARKER_RE.match(text) is not None



import builtins
import os.path
import re
import token

from turcar import assistance
from turcar.assistance import (
    ErrorHelper,
    HelperNotSupportedError,
    Suggestion,
    add_error_helper,
    name_similarity,
)
from turcar.misc_utils import running_on_windows


class SyntaxErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        import tokenize

        super().__init__(error_info)

        self.tokens = []
        self.token_error = None

        if self.error_info["message"] == "EOL while scanning string literal":
            self.intro_text = (
                "You haven't properly closed the string on line %s." % self.error_info["lineno"]
                + "\n(If you want a multi-line string, then surround it with"
                + " `'''` or `\"\"\"` at both ends.)"
            )

        elif self.error_info["message"] == "EOF while scanning triple-quoted string literal":
            # lineno is not useful, as it is at the end of the file and user probably
            # didn't want the string to end there
            self.intro_text = "You haven't properly closed a triple-quoted string"

        else:
            if self.error_info["filename"] and os.path.isfile(self.error_info["filename"]):
                with open(self.error_info["filename"], mode="rb") as fp:
                    try:
                        for t in tokenize.tokenize(fp.readline):
                            self.tokens.append(t)
                    except tokenize.TokenError as e:
                        self.token_error = e
                    except IndentationError as e:
                        self.indentation_error = e

                if not self.tokens or self.tokens[-1].type not in [
                    token.ERRORTOKEN,
                    token.ENDMARKER,
                ]:
                    self.tokens.append(tokenize.TokenInfo(token.ERRORTOKEN, "", None, None, ""))
            else:
                self.tokens = []

            unbalanced = self._sug_unbalanced_parens()
            if unbalanced:
                self.intro_text = (
                    "Unbalanced parentheses, brackets or braces:\n\n" + unbalanced.body
                )
                self.intro_confidence = 5
            else:
                self.intro_text = "Python doesn't know how to read your program."

                if "^" in str(self.error_info):
                    self.intro_text += (
                        "\n\nSmall `^` in the original error message shows where it gave up,"
                        + " but the actual mistake can be before this."
                    )

                self.suggestions = [self._sug_missing_or_misplaced_colon()]

    def _sug_missing_or_misplaced_colon(self):
        import tokenize

        i = 0
        title = "Did you forget the colon?"
        relevance = 0
        body = ""
        while i < len(self.tokens) and self.tokens[i].type != token.ENDMARKER:
            t = self.tokens[i]
            if t.string in [
                "if",
                "elif",
                "else",
                "while",
                "for",
                "with",
                "try",
                "except",
                "finally",
                "class",
                "def",
            ]:
                keyword_pos = i
                while (
                    self.tokens[i].type
                    not in [
                        token.NEWLINE,
                        token.ENDMARKER,
                        token.COLON,  # colon may be OP
                        token.RBRACE,
                    ]
                    and self.tokens[i].string != ":"
                ):
                    old_i = i
                    if self.tokens[i].string in "([{":
                        i = self._skip_braced_part(i)
                        assert i > old_i
                        if i == len(self.tokens):
                            return None
                    else:
                        i += 1

                if self.tokens[i].string != ":":
                    relevance = 9
                    body = "`%s` header must end with a colon." % t.string
                    break

                # Colon was present, but maybe it should have been right
                # after the keyword.
                if (
                    t.string in ["else", "try", "finally"]
                    and self.tokens[keyword_pos + 1].string != ":"
                ):
                    title = "Incorrect use of `%s`" % t.string
                    body = "Nothing is allowed between `%s` and colon." % t.string
                    relevance = 9
                    if (
                        self.tokens[keyword_pos + 1].type not in (token.NEWLINE, tokenize.COMMENT)
                        and t.string == "else"
                    ):
                        body = "If you want to specify a condition, then use `elif` or nested `if`."
                    break

            i += 1

        return Suggestion("missing-or-misplaced-colon", title, body, relevance)

    def _sug_unbalanced_parens(self):
        problem = self._find_first_braces_problem()
        if not problem:
            return None

        return Suggestion("missing-or-misplaced-colon", "Unbalanced brackets", problem[1], 8)

    def _sug_wrong_increment_op(self):
        pass

    def _sug_wrong_decrement_op(self):
        pass

    def _sug_wrong_comparison_op(self):
        pass

    def _sug_switched_assignment_sides(self):
        pass

    def _skip_braced_part(self, token_index):
        assert self.tokens[token_index].string in ["(", "[", "{"]
        level = 1
        token_index += 1
        while token_index < len(self.tokens):
            if self.tokens[token_index].string in ["(", "[", "{"]:
                level += 1
            elif self.tokens[token_index].string in [")", "]", "}"]:
                level -= 1

            token_index += 1

            if level <= 0:
                return token_index

        assert token_index == len(self.tokens)
        return token_index

    def _find_first_braces_problem(self):
        # closers = {'(':')', '{':'}', '[':']'}
        openers = {")": "(", "}": "{", "]": "["}

        brace_stack = []
        for t in self.tokens:
            if t.string in ["(", "[", "{"]:
                brace_stack.append(t)
            elif t.string in [")", "]", "}"]:
                if not brace_stack:
                    return (
                        t,
                        "Found '`%s`' at `line %d <%s>`_ without preceding matching '`%s`'"
                        % (
                            t.string,
                            t.start[0],
                            assistance.format_file_url(
                                self.error_info["filename"], t.start[0], t.start[1]
                            ),
                            openers[t.string],
                        ),
                    )
                elif brace_stack[-1].string != openers[t.string]:
                    return (
                        t,
                        "Found '`%s`' at `line %d <%s>`__ when last unmatched opener was '`%s`' at `line %d <%s>`__"
                        % (
                            t.string,
                            t.start[0],
                            assistance.format_file_url(
                                self.error_info["filename"], t.start[0], t.start[1]
                            ),
                            brace_stack[-1].string,
                            brace_stack[-1].start[0],
                            assistance.format_file_url(
                                self.error_info["filename"],
                                brace_stack[-1].start[0],
                                brace_stack[-1].start[1],
                            ),
                        ),
                    )
                else:
                    brace_stack.pop()

        if brace_stack:
            return (
                brace_stack[-1],
                "'`%s`' at `line %d <%s>`_ is not closed by the end of the program"
                % (
                    brace_stack[-1].string,
                    brace_stack[-1].start[0],
                    assistance.format_file_url(
                        self.error_info["filename"],
                        brace_stack[-1].start[0],
                        brace_stack[-1].start[1],
                    ),
                ),
            )

        return None


class NameErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)

        names = re.findall(r"\'.*\'", error_info["message"])
        assert len(names) == 1
        self.name = names[0].strip("'")

        self.intro_text = "Python doesn't know what `%s` stands for." % self.name
        self.suggestions = [
            self._sug_bad_spelling(),
            self._sug_missing_quotes(),
            self._sug_missing_import(),
            self._sug_local_from_global(),
            self._sug_not_defined_yet(),
        ]

    def _sug_missing_quotes(self):
        if self._is_attribute_value() or self._is_call_function() or self._is_subscript_value():
            relevance = 0
        else:
            relevance = 5

        return Suggestion(
            "missing-quotes",
            "Did you actually mean string (text)?",
            'If you didn\'t mean a variable but literal text "%s", then surround it with quotes.'
            % self.name,
            relevance,
        )

    def _sug_bad_spelling(self):
        # Yes, it would be more proper to consult builtins from the backend,
        # but it's easier this way...
        all_names = {name for name in dir(builtins) if not name.startswith("_")}
        all_names |= {"pass", "break", "continue", "return", "yield"}

        if self.last_frame.globals is not None:
            all_names |= set(self.last_frame.globals.keys())
        if self.last_frame.locals is not None:
            all_names |= set(self.last_frame.locals.keys())

        similar_names = {self.name}
        if all_names:
            relevance = 0
            for name in all_names:
                sim = name_similarity(name, self.name)
                if sim > 4:
                    similar_names.add(name)
                relevance = max(sim, relevance)
        else:
            relevance = 3

        if len(similar_names) > 1:
            body = "I found similar names. Are all of them spelled correctly?\n\n"
            for name in sorted(similar_names, key=lambda x: x.lower()):
                # TODO: add location info
                body += "* `%s`\n\n" % name
        else:
            body = (
                "Compare the name with corresponding definition / assignment / documentation."
                + " Don't forget that case of the letters matters!"
            )

        return Suggestion("bad-spelling-name", "Did you misspell it (somewhere)?", body, relevance)

    def _sug_missing_import(self):
        likely_importable_functions = {
            "math": {"ceil", "floor", "sqrt", "sin", "cos", "degrees"},
            "random": {"randint"},
            "turtle": {
                "left",
                "right",
                "forward",
                "fd",
                "goto",
                "setpos",
                "Turtle",
                "penup",
                "up",
                "pendown",
                "down",
                "color",
                "pencolor",
                "fillcolor",
                "begin_fill",
                "end_fill",
                "pensize",
                "width",
            },
            "re": {"search", "match", "findall"},
            "datetime": {"date", "time", "datetime", "today"},
            "statistics": {
                "mean",
                "median",
                "median_low",
                "median_high",
                "mode",
                "pstdev",
                "pvariance",
                "stdev",
                "variance",
            },
            "os": {"listdir"},
            "time": {"time", "sleep"},
        }

        body = None

        if self._is_call_function():
            relevance = 5
            for mod in likely_importable_functions:
                if self.name in likely_importable_functions[mod]:
                    relevance += 3
                    body = (
                        "If you meant `%s` from module `%s`, then add\n\n`from %s import %s`\n\nto the beginning of your script."
                        % (self.name, mod, mod, self.name)
                    )
                    break

        elif self._is_attribute_value():
            relevance = 5
            body = (
                "If you meant module `%s`, then add `import %s` to the beginning of your script"
                % (self.name, self.name)
            )

            if self.name in likely_importable_functions:
                relevance += 3

        elif self._is_subscript_value() and self.name != "argv":
            relevance = 0
        elif self.name == "pi":
            body = "If you meant the constant π, then add `from math import pi` to the beginning of your script."
            relevance = 8
        elif self.name == "argv":
            body = "If you meant the list with program arguments, then add `from sys import argv` to the beginning of your script."
            relevance = 8
        else:
            relevance = 3

        if body is None:
            body = "Some functions/variables need to be imported before they can be used."

        return Suggestion("missing-import", "Did you forget to import it?", body, relevance)

    def _sug_local_from_global(self):
        import ast

        relevance = 0
        body = None

        if self.last_frame.code_name == "<module>" and self.last_frame_module_ast is not None:
            function_names = set()
            for node in ast.walk(self.last_frame_module_ast):
                if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                    if self.name in map(lambda x: x.arg, node.args.args):
                        function_names.add(node.name)
                    # TODO: varargs, kw, ...
                    declared_global = False
                    for localnode in ast.walk(node):
                        # print(node.name, localnode)
                        if (
                            isinstance(localnode, ast.Name)
                            and localnode.id == self.name
                            and isinstance(localnode.ctx, ast.Store)
                        ):
                            function_names.add(node.name)
                        elif isinstance(localnode, ast.Global) and self.name in localnode.names:
                            declared_global = True

                    if node.name in function_names and declared_global:
                        function_names.remove(node.name)

            if function_names:
                relevance = 9
                body = (
                    (
                        "Name `%s` defined in `%s` is not accessible in the global/module level."
                        % (self.name, " and ".join(function_names))
                    )
                    + "\n\nIf you need that data at the global level, then consider changing the function so that it `return`-s the value."
                )

        return Suggestion(
            "local-from-global",
            "Are you trying to access a local variable outside of the function?",
            body,
            relevance,
        )

    def _sug_not_defined_yet(self):
        return Suggestion(
            "not-defined-yet",
            "Has Python executed the definition?",
            (
                "Don't forget that name becomes defined when corresponding definition ('=', 'def' or 'import') gets executed."
                + " If the definition comes later in code or is inside an if-statement, Python may not have executed it (yet)."
                + "\n\n"
                + "Make sure Python arrives to the definition before it arrives to this line. When in doubt, "
                + "`use the debugger <debuggers.rst>`_."
            ),
            2,
        )

    def _sug_maybe_attribute(self):
        "TODO:"

    def _sug_synonym(self):
        "TODO:"

    def _is_call_function(self):
        return self.name + "(" in (
            self.error_info["line"].replace(" ", "").replace("\n", "").replace("\r", "")
        )

    def _is_subscript_value(self):
        return self.name + "[" in (
            self.error_info["line"].replace(" ", "").replace("\n", "").replace("\r", "")
        )

    def _is_attribute_value(self):
        return self.name + "." in (
            self.error_info["line"].replace(" ", "").replace("\n", "").replace("\r", "")
        )


class AttributeErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)

        names = re.findall(r"\'.*?\'", error_info["message"])
        if len(names) != 2:
            # Can happen eg. in PGZero
            # https://github.com/turcar/turcar/issues/1535
            raise HelperNotSupportedError()

        self.type_name = names[0].strip("'")
        self.att_name = names[1].strip("'")

        self.intro_text = (
            "Your program tries to "
            + ("call method " if self._is_call_function() else "access attribute ")
            + "`%s` of " % self.att_name
            + _get_phrase_for_object(self.type_name)
            + ", but this type doesn't have such "
            + ("method." if self._is_call_function() else "attribute.")
        )

        self.suggestions = [
            self._sug_wrong_attribute_instead_of_len(),
            self._sug_bad_spelling(),
            self._sug_bad_type(),
        ]

    def _sug_wrong_attribute_instead_of_len(self):
        if self.type_name == "str":
            goal = "length"
        elif self.type_name == "bytes":
            goal = "number of bytes"
        elif self.type_name == "list":
            goal = "number of elements"
        elif self.type_name == "tuple":
            goal = "number of elements"
        elif self.type_name == "set":
            goal = "number of elements"
        elif self.type_name == "dict":
            goal = "number of entries"
        else:
            return None

        return Suggestion(
            "wrong-attribute-instead-of-len",
            "Did you mean to ask the %s?" % goal,
            "This can be done with function `len`, eg:\n\n`len(%s)`"
            % _get_sample_for_type(self.type_name),
            (9 if self.att_name.lower() in ("len", "length", "size") else 0),
        )

    def _sug_bad_spelling(self):
        # TODO: compare with attributes of known types
        return Suggestion(
            "bad-spelling-attribute",
            "Did you misspell the name?",
            "Don't forget that case of the letters matters too!",
            3,
        )

    def _sug_bad_type(self):
        if self._is_call_function():
            action = "call this function on"
        else:
            action = "ask this attribute from"

        return Suggestion(
            "wrong-type-attribute",
            "Did you expect another type?",
            "If you didn't mean %s %s, " % (action, _get_phrase_for_object(self.type_name))
            + "then step through your program to see "
            + "why this type appears here.",
            3,
        )

    def _is_call_function(self):
        return "." + self.att_name + "(" in (
            self.error_info["line"].replace(" ", "").replace("\n", "").replace("\r", "")
        )


class OSErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)

        if "Address already in use" in self.error_info["message"]:
            self.intro_text = "Your programs tries to listen on a port which is already taken."
            self.suggestions = [
                Suggestion(
                    "kill-by-port-type-error",
                    "Want to close the other process?",
                    self.get_kill_process_instructions(),
                    5,
                ),
                Suggestion(
                    "use-another-type-error",
                    "Can you use another port?",
                    "If you don't want to mess with the other process, then check whether"
                    + " you can configure your program to use another port.",
                    3,
                ),
            ]

        else:
            self.intro_text = "No specific information is available for this error."

    def get_kill_process_instructions(self):
        s = (
            "Let's say you need port 5000. If you don't know which process is using it,"
            + " then enter following system command into turcar's Shell:\n\n"
        )

        if running_on_windows():
            s += (
                "``!netstat -ano | findstr :5000``\n\n"
                + "You should see the process ID in the last column.\n\n"
            )
        else:
            s += (
                "``!lsof -i:5000``\n\n" + "You should see the process ID under the heading PID.\n\n"
            )

        s += (
            "Let's pretend the ID is 12345."
            " You can try hard-killing the process with following command:\n\n"
        )

        if running_on_windows():
            s += "``!tskill 12345``\n"
        else:
            s += (
                "``!kill -9 12345``\n\n"
                + "Both steps can be combined into single command:\n\n"
                + "``!kill -9 $(lsof -t -i:5000)``\n\n"
            )

        return s


class TypeErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)

        self.intro_text = (
            "Python was asked to do an operation with an object which " + "doesn't support it."
        )

        self.suggestions = [
            Suggestion(
                "step-to-find-type-error",
                "Did you expect another type?",
                "Step through your program to see why this type appears here.",
                3,
            ),
            Suggestion(
                "look-documentation-type-error",
                "Maybe you forgot some details about this operation?",
                "Look up the documentation or perform a web search with the error message.",
                2,
            ),
        ]

        # overwrite / add for special cases
        # something + str or str + something
        for r, string_first in [
            (r"unsupported operand type\(s\) for \+: '(.+?)' and 'str'", False),
            (r"^Can't convert '(.+?)' object to str implicitly$", True),  # Python 3.5
            (r"^must be str, not (.+)$", True),  # Python 3.6
            (r'^can only concatenate str (not "(.+?)") to str$', True),  # Python 3.7
        ]:
            m = re.match(r, error_info["message"], re.I)  # @UndefinedVariable
            if m is not None:
                self._bad_string_concatenation(m.group(1), string_first)
                return

        # TODO: other operations, when one side is string

    def _bad_string_concatenation(self, other_type_name, string_first):
        self.intro_text = "Your program is trying to put together " + (
            "a string and %s." if string_first else "%s and a string."
        ) % _get_phrase_for_object(other_type_name)

        self.suggestions.append(
            Suggestion(
                "convert-other-operand-to-string",
                "Did you mean to treat both sides as text and produce a string?",
                "In this case you should apply function `str` to the %s "
                % _get_phrase_for_object(other_type_name, False)
                + "in order to convert it to string first, eg:\n\n"
                + ("`'abc' + str(%s)`" if string_first else "`str(%s) + 'abc'`")
                % _get_sample_for_type(other_type_name),
                8,
            )
        )

        if other_type_name in ("float", "int"):
            self.suggestions.append(
                Suggestion(
                    "convert-other-operand-to-number",
                    "Did you mean to treat both sides as numbers and produce a sum?",
                    "In this case you should first convert the string to a number "
                    + "using either function `float` or `int`, eg:\n\n"
                    + ("`float('3.14') + 22`" if string_first else "`22 + float('3.14')`"),
                    7,
                )
            )


def _get_phrase_for_object(type_name, with_article=True):
    friendly_names = {
        "str": "a string",
        "int": "an integer",
        "float": "a float",
        "list": "a list",
        "tuple": "a tuple",
        "dict": "a dictionary",
        "set": "a set",
        "bool": "a boolean",
    }
    result = friendly_names.get(type_name, "an object of type '%s'" % type_name)

    if with_article:
        return result
    else:
        _, rest = result.split(" ", maxsplit=1)
        return rest


def _get_sample_for_type(type_name):
    if type_name == "int":
        return "42"
    elif type_name == "float":
        return "3.14"
    elif type_name == "str":
        return "'abc'"
    elif type_name == "bytes":
        return "b'abc'"
    elif type_name == "list":
        return "[1, 2, 3]"
    elif type_name == "tuple":
        return "(1, 2, 3)"
    elif type_name == "set":
        return "{1, 2, 3}"
    elif type_name == "dict":
        return "{1 : 'one', 2 : 'two'}"
    else:
        return "..."


def load_plugin():
    for name in globals():
        if name.endswith("ErrorHelper") and not name.startswith("_"):
            type_name = name[: -len("Helper")]
            add_error_helper(type_name, globals()[name])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from typing import Final, List, NoReturn, Optional, Tuple, Union, overload

import audio  # this module is available via microbit since V2

def panic(n: Optional[int]=...) -> NoReturn:
    """
    Enter a panic mode. Requires restart. Pass in an arbitrary integer <= 255 to indicate a status.
    """
    pass


def sleep(time: Union[int, float]) -> None:
    """
    Wait for n milliseconds. One second is 1000 milliseconds, so:

        microbit.sleep(1000)

    will pause the execution for one second. n can be an integer or a floating point number.
    """
    pass

def reset() -> NoReturn:
    """
    Restart the board.
    """

def running_time() -> int:
    """
    Return the number of milliseconds since the board was switched on or restarted.
    """
    pass


def temperature() -> int:
    """
    Return the temperature of the micro:bit in degrees Celcius.
    """
    pass

# Image
class Image(object):
    @overload
    def __init__(self, image: str) -> None:
        """
        String must contain digits 0-9 arranged into lines, describing the image, for example:

            image = Image("90009:"
                          "09090:"
                          "00900:"
                          "09090:"
                          "90009")

        will create a 5×5 image of an X. The end of a line is indicated by a colon.

        It’s also possible to use a newline (n) to indicate the end of a line like this:

            image = Image("90009\n"
                          "09090\n"
                          "00900\n"
                          "09090\n"
                          "90009")
        """

    @overload
    def __init__(self, width=None, height=None, buffer=None) -> None:
        """
        Create empty image with width columns and height rows. Optionally buffer can be an array
        of width x height integers in range 0-9 to initialize the image:

            Image(2, 2, b'\x08\x08\x08\x08')

        or:

            Image(2, 2, bytearray([9,9,9,9]))

        Will create a 2 x 2 pixel image at full brightness.
        """

    def width(self) -> int:
        """
        Return the width of the image in pixels.
        """
        pass

    def height(self) -> int:
        """
        Return the height of the image in pixels.
        """
        pass

    def get_pixel(self, x: int, y: int) -> int:
        """
        Return the brightness of pixel at column x and row y as an integer between 0 and 9.
        """
        pass

    def set_pixel(self, x: int, y: int, b: int) -> None:
        """
        Set the brightness of the pixel at column x and row y to the value, which has to be between
        0 (dark) and 9 (bright).

        This method will raise an exception when called on any of the built-in read-only images,
        like Image.HEART.
        """
        pass

    def shift_left(self, n: int) -> 'Image':
        """
        Return a new image created by shifting the picture left by n columns.
        """
        pass

    def shift_right(self, n: int) -> 'Image':
        """
        Same as image.shift_left(-n).
        """
        pass

    def shift_up(self, n: int) -> 'Image':
        """
        Return a new image created by shifting the picture up by n rows.
        """
        pass

    def shift_down(self, n: int) -> 'Image':
        """
        Same as image.shift_up(-n).
        """
        pass

    def copy(self) -> 'Image':
        """
        Return an exact copy of the image.
        """
        pass

    def crop(self, x1: int, y1: int, x2: int, y2: int) -> 'Image':
        """
        Return a new image by cropping the picture to a width of w and a height of h, starting with
        the pixel at column x and row y.
        """
        pass

    def invert(self) -> 'Image':
        """
        Return a new image by inverting the brightness of the pixels in the source image.
        """
        pass

    def fill(self, value) -> None:
        """
        Set the brightness of all the pixels in the image to the value, which has to be between 0
        (dark) and 9 (bright).

        This method will raise an exception when called on any of the built-in read-only images,
        like Image.HEART.
        """

    def blit(self, src: "Image", x: int, y: int, w: int, h: int, xdest=0, ydest=0) -> None:
        """
        Copy the rectangle defined by x, y, w, h from the image src into this image at xdest, ydest.
        Areas in the source rectangle, but outside the source image are treated as having a value of 0.

        shift_left(), shift_right(), shift_up(), shift_down() and crop() can are all implemented by using blit(). For example, img.crop(x, y, w, h) can be implemented as:

            def crop(self, x, y, w, h):
                res = Image(w, h)
                res.blit(self, x, y, w, h)
                return res
        """

    def __add__(self, other: "Image") -> "Image":
        """Create a new image by adding the brightness values from the two images for each pixel."""

    def __mul__(self, n: int) -> "Image":
        """
        Create a new image by multiplying the brightness of each pixel by n.
        """

    HEART : "Image" = ...
    HEART_SMALL : "Image" = ...
    HAPPY : "Image" = ...
    SMILE : "Image" = ...
    SAD : "Image" = ...
    CONFUSED : "Image" = ...
    ANGRY : "Image" = ...
    ASLEEP : "Image" = ...
    SURPRISED : "Image" = ...
    SILLY : "Image" = ...
    FABULOUS : "Image" = ...
    MEH : "Image" = ...
    YES : "Image" = ...
    NO : "Image" = ...
    CLOCK12 : "Image" = ...
    CLOCK11 : "Image" = ...
    CLOCK10 : "Image" = ...
    CLOCK9 : "Image" = ...
    CLOCK8 : "Image" = ...
    CLOCK7 : "Image" = ...
    CLOCK6 : "Image" = ...
    CLOCK5 : "Image" = ...
    CLOCK4 : "Image" = ...
    CLOCK3 : "Image" = ...
    CLOCK2 : "Image" = ...
    CLOCK1 : "Image" = ...
    ARROW_N : "Image" = ...
    ARROW_NE : "Image" = ...
    ARROW_E : "Image" = ...
    ARROW_SE : "Image" = ...
    ARROW_S : "Image" = ...
    ARROW_SW : "Image" = ...
    ARROW_W : "Image" = ...
    ARROW_NW : "Image" = ...
    TRIANGLE : "Image" = ...
    TRIANGLE_LEFT : "Image" = ...
    CHESSBOARD : "Image" = ...
    DIAMOND : "Image" = ...
    DIAMOND_SMALL : "Image" = ...
    SQUARE : "Image" = ...
    SQUARE_SMALL : "Image" = ...
    RABBIT : "Image" = ...
    COW : "Image" = ...
    MUSIC_CROTCHET : "Image" = ...
    MUSIC_QUAVER : "Image" = ...
    MUSIC_QUAVERS : "Image" = ...
    PITCHFORK : "Image" = ...
    XMAS : "Image" = ...
    PACMAN : "Image" = ...
    TARGET : "Image" = ...
    TSHIRT : "Image" = ...
    ROLLERSKATE : "Image" = ...
    DUCK : "Image" = ...
    HOUSE : "Image" = ...
    TORTOISE : "Image" = ...
    BUTTERFLY : "Image" = ...
    STICKFIGURE : "Image" = ...
    GHOST : "Image" = ...
    SWORD : "Image" = ...
    GIRAFFE : "Image" = ...
    SKULL : "Image" = ...
    UMBRELLA : "Image" = ...
    SNAKE : "Image" = ...

    ALL_CLOCKS: List["Image"] = ...
    ALL_ARROWS: List["Image"] = ...




# Accelerometer 3D orientation
class _MicroBitAccelerometer(object):
    """
    This object gives you access to the on-board accelerometer. The accelerometer also provides
    convenience functions for detecting gestures. The recognised gestures are: up, down, left, right,
    face up, face down, freefall, 3g, 6g, 8g, shake.

    By default MicroPython sets the accelerometer range to +/- 2g, changing the accelerometer range
    is currently not possible in MicroPython. The accelerometer returns a value in the range 0..1024
    for each axis, which is then scaled accordingly.
    """
    def get_x(self) -> int:
        """
        Get the acceleration measurement (tilt) in the x axis, as a positive or negative integer, depending
        on the direction. The measurement is given in milli-g. By default the accelerometer is
        configured with a range of +/- 2g, and so this method will return within the range of +/- 2000mg.
        """
        pass

    def get_y(self) -> int:
        """
        Get the acceleration measurement (tilt) in the y axis, as a positive or negative integer, depending
        on the direction. The measurement is given in milli-g. By default the accelerometer is
        configured with a range of +/- 2g, and so this method will return within the range of +/- 2000mg.
        """
        pass

    def get_z(self) -> int:
        """
        Get the acceleration measurement in the z axis (up-down motion), as a positive or negative integer, depending
        on the direction. The measurement is given in milli-g. By default the accelerometer is
        configured with a range of +/- 2g, and so this method will return within the range of +/- 2000mg.
        """
        pass

    def get_values(self) -> Tuple[int, int, int]:
        """
        Get the acceleration measurements in all axes at once, as a three-element tuple of integers
        ordered as X, Y, Z. By default the accelerometer is configured with a range of +/- 2g, and
        so X, Y, and Z will be within the range of +/-2000mg.
        """

    def current_gesture(self) -> str:
        """
        Return the name of the current gesture.

        MicroPython understands the following gesture names: "up", "down", "left", "right", "face up",
        "face down", "freefall", "3g", "6g", "8g", "shake". Gestures are always represented as strings.

        No gesture is expressed as "".
        """

    def is_gesture(self, name: str) -> bool:
        """
        Return True or False to indicate if the named gesture is currently
        active.
        MicroPython understands the following gestures: 'up', 'down', 'left',
        'right', 'face up', 'face down', 'freefall', '3g', '6g', '8g' and
        'shake'.
        """
        pass

    def was_gesture(self, name: str) -> bool:
        """
        Return True or False to indicate if the named gesture was active since
        the last call.
        MicroPython understands the following gestures: 'up', 'down', 'left',
        'right', 'face up', 'face down', 'freefall', '3g', '6g', '8g' and
        'shake'.
        """
        pass

    def get_gestures(self) -> Tuple[str, ...]:
        """
        Return a tuple indicating the gesture history. The most recent gesture
        is last.
        Calling this method also clears the gesture history.
        MicroPython understands the following gestures: 'up', 'down', 'left',
        'right', 'face up', 'face down', 'freefall', '3g', '6g', '8g' and
        'shake'.
        """
        pass


accelerometer : _MicroBitAccelerometer = ...


# Pushbutton
class _MicroBitButton(object):
    def is_pressed(self) -> bool:
        """
        Returns True if the this button is currently being held down, and False otherwise.
        """
        pass

    def was_pressed(self) -> bool:
        """
        Returns True or False to indicate if the button was pressed (went from up to down) since the
        device started or the last time this method was called. Calling this method will clear the
        press state so that the button must be pressed again before this method will return True again.
        """
        pass

    def get_presses(self) -> int:
        """
        Returns the running total of button presses, and resets this total to zero before returning.
        """
        pass


button_a : _MicroBitButton = ...
button_b : _MicroBitButton = ...


# Compass 3D direction heading
class _MicroBitCompass(object):
    def is_calibrated(self) -> bool:
        """
        Returns True if the compass has been successfully calibrated, and returns False otherwise.
        """
        pass

    def calibrate(self) -> None:
        """
        Starts the calibration process. An instructive message will be scrolled to the user after
        which they will need to rotate the device in order to draw a circle on the LED display.
        """
        pass

    def clear_calibration(self) -> None:
        """
        Undoes the calibration, making the compass uncalibrated again.
        """
        pass

    def get_x(self) -> int:
        """
        Gives the reading of the magnetic field strength on the x axis in nano tesla, as a positive
        or negative integer, depending on the direction of the field.
        """
        pass

    def get_y(self) -> int:
        """
        Gives the reading of the magnetic field strength on the y axis in nano tesla, as a positive
        or negative integer, depending on the direction of the field.
        """
        pass

    def get_z(self) -> int:
        """
        Gives the reading of the magnetic field strength on the z axis in nano tesla, as a positive
        or negative integer, depending on the direction of the field.
        """
        pass

    def get_field_strength(self) -> int:
        """
        Returns an integer indication of the magnitude of the magnetic field around the device in nano tesla.
        """
        pass

    def heading(self) -> int:
        """
        Gives the compass heading, calculated from the above readings, as an integer in the range
        from 0 to 360, representing the angle in degrees, clockwise, with north as 0.
        """
        pass


compass : _MicroBitCompass = ...


# Display 5x5 LED grid
class _MicroBitDisplay(object):
    """
    This module controls the 5×5 LED display on the front of your board. It can be used to display
    images, animations and even text.
    """

    def get_pixel(self, x: int, y: int) -> int:
        """
        Return the brightness of the LED at column x and row y as an integer between 0 (off) and 9 (bright).
        """
        pass

    def set_pixel(self, x: int, y: int, value: int) -> None:
        """
        Set the brightness of the LED at column x and row y to value, which has to be an integer between 0 and 9.
        """
        pass

    def clear(self) -> None:
        """
        Set the brightness of all LEDs to 0 (off).
        """
        pass

    @overload
    def show(self, image: Image) -> None:
        """
        Display the image.
        """

    @overload
    def show(
        self,
        value: Union[int, float, str, 'Image', List['Image']],
        delay: int = 400,
        *,
        wait: bool = True,
        loop: bool = False,
        clear: bool = False,
    ) -> None:
        """
        If value is a string, float or integer, display letters/digits in sequence. Otherwise, if
        value is an iterable sequence of images, display these images in sequence. Each letter,
        digit or image is shown with delay milliseconds between them.

        If wait is True, this function will block until the animation is finished, otherwise the
        animation will happen in the background.

        If loop is True, the animation will repeat forever.

        If clear is True, the display will be cleared after the iterable has finished.

        Note that the wait, loop and clear arguments must be specified using their keyword.

        Note. If using a generator as the iterable, then take care not to allocate any memory in the
        generator as allocating memory in an interrupt is prohibited and will raise a MemoryError.
        """
        pass

    def scroll(
        self,
        value: str,
        delay: int = 150,
        *,
        wait: bool = True,
        loop: bool = False,
        monospace: bool = False,
    ) -> None:
        """
        Scrolls value horizontally on the display. If value is an integer or float it is first
        converted to a string using str(). The delay parameter controls how fast the text is scrolling.

        If wait is True, this function will block until the animation is finished, otherwise the
        animation will happen in the background.

        If loop is True, the animation will repeat forever.

        If monospace is True, the characters will all take up 5 pixel-columns in width, otherwise
        there will be exactly 1 blank pixel-column between each character as they scroll.

        Note that the wait, loop and monospace arguments must be specified using their keyword.
        """
        pass

    def on(self) -> None:
        """
        Turn on the display.
        """
        pass

    def off(self) -> None:
        """
        Turn off the display (thus allowing you to re-use the GPIO pins associated with the display
        for other purposes).
        """
        pass

    def is_on(self) -> bool:
        """
        Returns True if the display is on, otherwise returns False.
        """
        pass

    def read_light_level(self) -> int:
        """
        Use the display’s LEDs in reverse-bias mode to sense the amount of light falling on the display.
        Returns an integer between 0 and 255 representing the light level, with larger meaning more light.
        """

display : _MicroBitDisplay = ...


# Pins
class _MicroBitDigitalPin(object):
    """
    A standard pin
    """

    PULL_UP: Final[int] = 0
    PULL_DOWN: Final[int] = 1
    NO_PULL: Final[int] = 2


    def read_digital(self) -> int:
        """
        Read the digital value of the pin. 0 for low, 1 for high
        """
        pass

    def write_digital(self, value: int) -> None:
        """
        Set the pin to output high if value is 1, or to low, it it is 0.
        """
        pass

    def set_pull(self, mode: int) -> None:
        """
        Set the pull state to one of three possible values: pin.PULL_UP, pin.PULL_DOWN or pin.NO_PULL
        (where pin is an instance of a pin).
        """

    def get_pull(self) -> int:
        """
        Returns the pull configuration on a pin, which can be one of three possible values: NO_PULL,
        PULL_DOWN, or PULL_UP. These are set using the set_pull() method or automatically configured
        when a pin mode requires it.
        """

    def get_mode(self) -> str:
        """
        Returns the pin mode. When a pin is used for a specific function, like writing a digital
        value, or reading an analog value, the pin mode changes. Pins can have one of the following
        modes: "unused", "analog", "read_digital", "write_digital", "display", "button", "music",
        "audio", "touch", "i2c", "spi".
        """


class _MicroBitAnalogDigitalPin(_MicroBitDigitalPin):
    """
    These pins have ADC & PWM support
    """
    def read_analog(self) -> int:
        """
        Read the voltage applied to the pin.
        Returns the reading as a number between 0 (meaning 0V) and 1023
        (meaning 3.3V).
        """
        pass

    def write_analog(self, value: int) -> None:
        """
        Output a PWM signal on the pin, with the duty cycle proportional to the
        provided value. The value may be either an integer or a floating point number
        between 0 (0% duty cycle) and 1023 (100% duty).
        """
        pass

    def set_analog_period(self, period: int) -> None:
        """
        Set the period of the PWM signal output to period milliseconds.
        The minimum valid value is 1ms.
        """
        pass

    def set_analog_period_microseconds(self, period: int) -> None:
        """
        Set the period of the PWM signal output to period microseconds.
        The minimum valid value is 256µs.
        """
        pass

    def get_analog_period_microseconds(self) -> int:
        ...

class _MicroBitTouchPinBase:
    RESISTIVE: Final[int] = 0
    """Constant for set_touch_mode(...). Available since V2"""
    CAPACITIVE: Final[int] = 1
    """Constant for set_touch_mode(...). Available since V2"""

    def is_touched(self) -> bool:
        """
        Return True if the pin is being touched with a finger, otherwise return False.

        Resistive touch. This test is done by measuring how much resistance there is between the pin
        and ground. A low resistance gives a reading of True. To get a reliable reading using a
        finger you may need to touch the ground pin with another part of your body, for example your
        other hand.

        Capacitive touch. This test is done by interacting with the electric field of a capacitor
        using a finger as a conductor. Capacitive touch does not require you to make a ground
        connection as part of a circuit.
        """
        pass

    def set_touch_mode(self, mode:int) -> None:
        """
        Available since V2. Use pin.CAPACITIVE (no need to touch GND with other hand)
        or pin.RESISTIVE as argument.

        The default for the edge connector pins is resistive and the logo pin in V2 is capacitive.
        """


class _MicroBitTouchPin(_MicroBitTouchPinBase, _MicroBitAnalogDigitalPin):
    """
    These pins are aranged on the bottom of the board and have holes through them
    """

class _MicroBitTouchOnlyPin(_MicroBitTouchPinBase):
    """
    A touch sensitive logo pin on the front of the micro:bit, which by default is set to capacitive touch mode.
    """

pin0 : _MicroBitTouchPin = ...
pin1 : _MicroBitTouchPin = ...
pin2 : _MicroBitTouchPin = ...
pin3 : _MicroBitAnalogDigitalPin = ...
pin4 : _MicroBitAnalogDigitalPin = ...
pin5 : _MicroBitDigitalPin = ...
pin6 : _MicroBitDigitalPin = ...
pin7 : _MicroBitDigitalPin = ...
pin8 : _MicroBitDigitalPin = ...
pin9 : _MicroBitDigitalPin = ...
pin10 : _MicroBitAnalogDigitalPin = ...
pin11 : _MicroBitDigitalPin = ...
pin12 : _MicroBitDigitalPin = ...
pin13 : _MicroBitDigitalPin = ...
pin14 : _MicroBitDigitalPin = ...
pin15 : _MicroBitDigitalPin = ...
pin16 : _MicroBitDigitalPin = ...
# Pin17 : 3v3
# Pin18 : 3v3
pin19 : _MicroBitDigitalPin = ...
pin20 : _MicroBitDigitalPin = ...
# pin21 : gnd
# pin22 : gnd

pin_logo : _MicroBitTouchOnlyPin = ...
"""A touch sensitive logo pin on the front of the micro:bit, which by default is set to capacitive touch mode."""
pin_speaker: _MicroBitDigitalPin = ...
"""A pin to address the micro:bit speaker. This API is intended only for use in Pulse-Width Modulation pin operations eg. pin_speaker.write_analog(128)."""

# I2C
class _MicroBitI2C(object):
    """
    The i2c module lets you communicate with devices connected to your board using the I²C bus protocol.
    There can be multiple slave devices connected at the same time, and each one has its own unique
    address, that is either fixed for the device or configured on it. Your board acts as the I²C master.

    We use 7-bit addressing for devices because of the reasons stated here.

    This may be different to other micro:bit related solutions.

    How exactly you should communicate with the devices, that is, what bytes to send and how to
    interpret the responses, depends on the device in question and should be described separately in
    that device’s documentation.
    """

    def init(self, frequency: int=100000, sda: _MicroBitDigitalPin=pin20,
             scl: _MicroBitDigitalPin=pin19) -> None:
        """
        Re-initialize peripheral with the specified clock frequency freq on the specified sda and scl pins.

        Warning: On a micro:bit V1 board, changing the I²C pins from defaults will make the
        accelerometer and compass stop working, as they are connected internally to those pins. This
        warning does not apply to the V2 revision of the micro:bit as this has separate I²C lines for
        the motion sensors and the edge connector.
        """
        pass

    def scan(self) -> List[int]:
        """
        Scan the bus for devices. Returns a list of 7-bit addresses corresponding to those devices
        that responded to the scan.
        """

    def read(self, addr: int, n: int, repeat: bool = False) -> bytes:
        """
        Read n bytes from the device with 7-bit address addr. If repeat is True, no stop bit will be sent.
        """
        pass

    def write(self, add: int, buffer: bytes, repeat: bool = False) -> None:
        """
        Write bytes from buf to the device with 7-bit address addr. If repeat is True, no stop bit will be sent.
        """
        pass


i2c : _MicroBitI2C = ...


# uart
class _MicroBitUART(object):
    ODD: int = 1
    EVEN: int = 0

    def init(
        self,
        baudrate: int = 9600,
        bits: int = 8,
        parity: Optional[int] = None,
        stop: int = 1,
        *,
        tx: Optional[_MicroBitDigitalPin] = None,
        rx: Optional[_MicroBitDigitalPin] = None,
    ) -> None:
        """
        Initialize serial communication with the specified parameters on the specified tx and rx pins.
        Note that for correct communication, the parameters have to be the same on both communicating devices.

        Warning: Initializing the UART on external pins will cause the Python console on USB to
        become unaccessible, as it uses the same hardware. To bring the console back you must
        reinitialize the UART without passing anything for tx or rx (or passing None to these arguments).
        This means that calling uart.init(115200) is enough to restore the Python console.

        The baudrate defines the speed of communication. Common baud rates include:
        9600, 14400, 19200, 28800, 38400, 57600, 115200.

        The bits defines the size of bytes being transmitted, and the board only supports 8.
        The parity parameter defines how parity is checked, and it can be None, microbit.uart.ODD or
        microbit.uart.EVEN. The stop parameter tells the number of stop bits, and has to be 1 for
        this board.

        If tx and rx are not specified then the internal USB-UART TX/RX pins are used which connect
        to the USB serial converter on the micro:bit, thus connecting the UART to your PC. You can
        specify any other pins you want by passing the desired pin objects to the tx and rx parameters.

        When connecting the device, make sure you “cross” the wires – the TX pin on your board needs
        to be connected with the RX pin on the device, and the RX pin – with the TX pin on the device.
        Also make sure the ground pins of both devices are connected.
        """
        pass

    def any(self) -> bool:
        """
        Return True if any data is waiting, else False.
        """
        pass

    def read(self, nbytes: Optional[int]=...) -> Optional[bytes]:
        """
        Read bytes. If nbytes is specified then read at most that many bytes, otherwise read as many
        bytes as possible.

        Return value: a bytes object or None on timeout.

        A bytes object contains a sequence of bytes. Because ASCII characters can fit in single bytes
        this type of object is often used to represent simple text and offers methods to manipulate
        it as such, e.g. you can display the text using the print() function.

        You can also convert this object into a string object, and if there are non-ASCII characters
        present the encoding can be specified:

            msg_bytes = uart.read()
            msg_str = str(msg, 'UTF-8')

        Note. The timeout for all UART reads depends on the baudrate and is otherwise not changeable
        via Python. The timeout, in milliseconds, is given by: microbit_uart_timeout_char = 13000 /
        baudrate + 1

        Note. The internal UART RX buffer is 64 bytes, so make sure data is read before the buffer
        is full or some of the data might be lost.

        Warning. Receiving 0x03 will stop your program by raising a Keyboard Interrupt. You can enable
        or disable this using micropython.kbd_intr().
        """
        pass


    def readinto(self, buf: bytes, nbytes: Optional[int]=None) -> Optional[int]:
        """
        Read bytes into the buf. If nbytes is specified then read at most that many bytes.
        Otherwise, read at most len(buf) bytes.

        Return value: number of bytes read and stored into buf or None on timeout.
        """
        pass

    def readline(self) -> Optional[bytes]:
        """
        Read a line, ending in a newline character.

        Return value: the line read or None on timeout. The newline character is included in the
        returned bytes.
        """

    def write(self, buf: Union[bytes, str]) -> int:
        """
        Write the buffer to the bus, it can be a bytes object or a string.
        """
        pass


uart : _MicroBitUART = ...


# SPI
class _MicroBitSPI(object):
    def init(
        self,
        baudrate: int = 1000000,
        bits: int = 8,
        mode: int = 0,
        sclk: _MicroBitDigitalPin = pin13,
        mosi: _MicroBitDigitalPin = pin15,
        miso: _MicroBitDigitalPin = pin14,
    ) -> None:
        """
        Initialize SPI communication with the specified parameters on the specified pins. Note that
        for correct communication, the parameters have to be the same on both communicating devices.

        The baudrate defines the speed of communication.

        The bits defines the size of bytes being transmitted. Currently only bits=8 is supported.
        However, this may change in the future.

        The mode determines the combination of clock polarity and phase according to the following
        convention, with polarity as the high order (leftmost) bit and phase as the low order bit:

            * 0 => 00
            * 1 => 01
            * 2 => 10
            * 3 => 11

        Polarity (aka CPOL) 0 means that the clock is at logic value 0 when idle and goes high
        (logic value 1) when active; polarity 1 means the clock is at logic value 1 when idle and
        goes low (logic value 0) when active. Phase (aka CPHA) 0 means that data is sampled on the
        leading edge of the clock, and 1 means on the trailing edge
        (viz. https://en.wikipedia.org/wiki/Signal_edge).

        The sclk, mosi and miso arguments specify the pins to use for each type of signal.
        """
        pass

    def write(self, buf: bytes) -> None:
        """
        Write the buffer of bytes to the bus.
        """
        pass

    def read(self, nbytes: int) -> bytes:
        """
        Read at most nbytes. Returns what was read.
        """
        pass

    def write_readinto(self, outbuf: bytes, inbuf: bytes) -> None:
        """
        Write the out buffer to the bus and read any response into the in buffer. The length of the
        buffers should be the same. The buffers can be the same object.
        """
        pass


spi : _MicroBitSPI = ...

class _MicroBitSpeaker:
    """
    Available since V2.
    """
    def off(self) -> None:
        """
        Turn off the speaker. This does not disable sound output to an edge connector pin.
        """
    def on(self) -> None:
        """
        Turn on the speaker.
        """

speaker: _MicroBitSpeaker = ...


class SoundEvent:
    """
    Enum for describing sound events. Available since V2.
    """
    LOUD: "SoundEvent" = ...
    QUIET: "SoundEvent" = ...

class _MicroBitMicrophone:
    """
    Available since V2.

    This object lets you access the built-in microphone available on the micro:bit V2. It can be used to respond to sound. The microphone input is located on the front of the board alongside a microphone activity LED, which is lit when the microphone is in use.

    The microphone can respond to a pre-defined set of sound events that are based on the amplitude and wavelength of the sound.

    These sound events are represented by instances of the SoundEvent class, accessible via variables in microbit.SoundEvent:

    microbit.SoundEvent.QUIET: Represents the transition of sound events, from loud to quiet like speaking or background music.
    microbit.SoundEvent.LOUD: Represents the transition of sound events, from quiet to loud like clapping or shouting.
    """

    def current_event(self) -> Optional[SoundEvent]:
        """
        return: the name of the last recorded sound event, SoundEvent('loud') or SoundEvent('quiet').
        """

    def was_event(self, event: SoundEvent) -> bool:
        """
        * event: a sound event, such as SoundEvent.LOUD or SoundEvent.QUIET.
        * return: true if sound was heard at least once since the last call, otherwise false. was_event() also clears the sound event history before returning.
        """
    def is_event(self, event: SoundEvent) -> bool:
        """
        * event: a sound event, such as SoundEvent.LOUD or SoundEvent.QUIET.
        * return: true if sound event is the most recent since the last call, otherwise false. It does not clear the sound event history.
        """

    def get_events(self) -> Tuple[SoundEvent]:
        """
        Return a tuple of the event history. The most recent is listed last. get_events() also clears the sound event history before returning.
        """

    def set_threshold(self, event: SoundEvent, value: int) -> None:
        """
        * event: a sound event, such as SoundEvent.LOUD or SoundEvent.QUIET.
        * value: The threshold level in the range 0-255. For example, set_threshold(SoundEvent.LOUD, 250) will only trigger if the sound is very loud (>= 250).
        """

    def sound_level(self) -> int:
        """
        Return a representation of the sound pressure level in the range 0 to 255.
        """

def set_volume(value: int) -> None:
    """
    sets the output volume (0-255) of the micro:bit speaker (V2) and
    external speaker or headphones connected to the edge connector pins.
    """

def ws2812_write(pin, buffer) -> None:
    ...

class Sound:
    """Enum for sound constants. Since V2. Can be used in audio.play"""

    GIGGLE: "Sound" = ...
    HAPPY: "Sound" = ...
    HELLO: "Sound" = ...
    MYSTERIOUS: "Sound" = ...
    SAD: "Sound" = ...
    SLIDE: "Sound" = ...
    SOARING: "Sound" = ...
    SPRING: "Sound" = ...
    TWINKLE: "Sound" = ...
    YAWN: "Sound" = ...
